"""
    picasso/lib
    ~~~~~~~~~~~~~~~~~~~~

    Handy functions and classes

    :author: Joerg Schnitzbauer, 2016
    :copyright: Copyright (c) 2016 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations
from collections.abc import Callable
from asyncio import Future as _Future
from typing import Any as _Any

import time as _time
import numba as _numba
import numpy as _np
from lmfit import Model as _Model
from numpy.lib.recfunctions import append_fields as _append_fields
from numpy.lib.recfunctions import drop_fields as _drop_fields
from numpy.lib.recfunctions import stack_arrays as _stack_arrays
import collections as _collections
import glob as _glob
import os.path as _ospath
from picasso import io as _io
from PyQt5 import QtCore, QtWidgets

# A global variable where we store all open progress and status dialogs.
# In case of an exception, we close them all,
# so that the GUI remains responsive.
_dialogs = []


class ProgressDialog(QtWidgets.QProgressDialog):
    """ProgressDialog displays a progress dialog with a progress bar."""

    def __init__(self, description, minimum, maximum, parent):
        # append time estimate to description
        super().__init__(
            description,
            None,
            minimum,
            maximum,
            parent,
            QtCore.Qt.CustomizeWindowHint,
        )
        self.description_base = description # without time estimate
        self.initalized = None

    def init(self):
        _dialogs.append(self)
        self.setMinimumDuration(500)
        self.setModal(True)
        self.t0 = _time.time()
        self.app = QtCore.QCoreApplication.instance()
        self.initalized = True

    def set_value(self, value):
        if not self.initalized:
            self.init()
        self.setValue(value)
        # estimate time left
        elapsed = _time.time() - self.t0
        remaining = int((self.maximum() - value) * elapsed / (value + 1e-6)) # s
        # convert to hh-mm-ss
        hours, remainder = divmod(remaining, 3600)
        minutes, seconds = divmod(remainder, 60)
        # format time estimate
        if hours > 0:
            hours = min(10, hours) # limit hours to 10 for display
            time_estimate = f"{hours:02d}h:{minutes:02d}m:{seconds:02d}s"
        else:
            time_estimate = f"{minutes:02d}m:{seconds:02d}s"
        # set label text with time estimate
        description = (
            f"{self.description_base}"
            f"\nEstimated time remaining: {time_estimate}"
        )
        self.setLabelText(description)
        self.app.processEvents()

    def closeEvent(self, event):
        _dialogs.remove(self)

    def zero_progress(self, description=None):
        """Sets progress dialog to zero and changes title if given."""

        if description:
            self.setLabelText(description)
        self.set_value(0)


class StatusDialog(QtWidgets.QDialog):
    """StatusDialog displays the description string in a dialog."""

    def __init__(self, description, parent):
        super(StatusDialog, self).__init__(parent, QtCore.Qt.CustomizeWindowHint)
        _dialogs.append(self)
        vbox = QtWidgets.QVBoxLayout(self)
        label = QtWidgets.QLabel(description)
        vbox.addWidget(label)
        self.show()
        QtCore.QCoreApplication.instance().processEvents()

    def closeEvent(self, event):
        _dialogs.remove(self)


class AutoDict(_collections.defaultdict):
    """A defaultdict whose auto-generated values are defaultdicts 
    itself. This allows for auto-generating nested values, e.g.
    a = AutoDict()
    a['foo']['bar']['carrot'] = 42
    """

    def __init__(self, *args, **kwargs):
        super().__init__(AutoDict, *args, **kwargs)


def cancel_dialogs():
    """Closes all open dialogs (ProgressDialog and StatusDialog) in the 
    GUI."""

    dialogs = [_ for _ in _dialogs]
    for dialog in dialogs:
        if isinstance(dialog, ProgressDialog):
            dialog.cancel()
        else:
            dialog.close()
    QtCore.QCoreApplication.instance().processEvents()  # just in case...


def cumulative_exponential(
    x: _np.ndarray, 
    a: float, 
    t: float, 
    c: float,
) -> _np.ndarray:
    """Used for binding kinetics estimation."""

    return a * (1 - _np.exp(-(x / t))) + c


CumulativeExponentialModel = _Model(cumulative_exponential)


def calculate_optimal_bins(
    data: _np.ndarray, 
    max_n_bins: int | None = None,
) -> _np.ndarray:
    """Calculates the optimal bins for display, for example, in 
    Picasso: Filter.
    
    Parameters
    ----------
    data : _np.ndarray
        Data to be binned.
    max_n_bins : int | None, optional
        Maximum number of bins.
    
    Returns
    -------
    bins : _np.ndarray
        Bins for display.
    """
    
    iqr = _np.subtract(*_np.percentile(data, [75, 25]))
    if iqr == 0:
        return _np.array([data[0] - 1.0, data[0] + 1.0])
    bin_size = 2 * iqr * len(data) ** (-1 / 3)
    if data.dtype.kind in ("u", "i") and bin_size < 1:
        bin_size = 1
    bin_min = data.min() - bin_size / 2
    try:
        n_bins = (data.max() - bin_min) / bin_size
        n_bins = int(n_bins)
    except:
        n_bins = 10
    if max_n_bins and n_bins > max_n_bins:
        n_bins = max_n_bins
    bins = _np.linspace(bin_min, data.max(), n_bins)
    return bins


def append_to_rec(
    rec_array: _np.recarray, 
    data: _np.ndarray, 
    name: str,
) -> _np.recarray:
    """Appends a new column to the existing _np.recarray.
    
    Parameters
    ----------
    rec_array : _np.recarray
        Recarray to which the new column is appended.
    data : _np.ndarray
        1D data to be appended.
    name : str
        Name of the new column.
    
    Returns
    -------
    rec_array : _np.recarray
        Recarray with the new column.
    """
    
    if hasattr(rec_array, name):
        rec_array = remove_from_rec(rec_array, name)
    rec_array = _append_fields(
        rec_array,
        name,
        data,
        dtypes=data.dtype,
        usemask=False,
        asrecarray=True,
    )
    return rec_array


def merge_locs(
    locs_list: list[_np.recarray], 
    increment_frames: bool = True,
) -> _np.recarray:
    """Merges localization lists into one file. Can increment frames
    to avoid overlapping frames.
    
    Parameters
    ----------
    locs_list : list of _np.recarrays
        List of localization lists to be merged.
    increment_frames : bool, optional
        If True, increments frames of each localization list by the
        maximum frame number of the previous localization list. Useful
        when the localization lists are from different movies but 
        represent the same stack. Default is True.
    
    Returns
    locs : _np.recarray
        Merged localizations.
    """

    if increment_frames:
        last_frame = 0
        for i, locs in enumerate(locs_list):
            locs["frame"] += last_frame
            last_frame = locs["frame"][-1].max()
            locs_list[i] = locs    
    locs = _stack_arrays(locs_list, usemask=False, asrecarray=True)
    return locs


def ensure_sanity(locs: _np.recarray, info: list[dict]) -> _np.recarray:
    """Ensures that localizations are within the image dimensions
    and have positive localization precisions and other parameters.
    
    Parameters
    ----------
    locs : _np.recarray
        Localizations list.
    info : list of dicts
        Localization metadata.
    
    Returns
    -------
    locs : _np.recarray
        Localizations that pass the sanity checks.
    """
    
    # no inf or nan:
    locs = locs[
        _np.all(
            _np.array([_np.isfinite(locs[_]) for _ in locs.dtype.names]),
            axis=0,
        )
    ]
    # other sanity checks:
    locs = locs[locs.x > 0]
    locs = locs[locs.y > 0]
    locs = locs[locs.x < info[0]["Width"]]
    locs = locs[locs.y < info[0]["Height"]]
    for attr in ["lpx", "lpy", "photons", "ellipticity", "sx", "sy"]:
        if hasattr(locs, attr):
            locs = locs[locs[attr] >= 0]
    return locs


def is_loc_at(x: float, y: float, locs: _np.recarray, r: float) -> _np.ndarray:
    """Checks which localizations are within radius r of position 
    (x, y).

    Parameters
    ----------
    x : float
        x-coordinate of the position.
    y : float
        y-coordinate of the position.
    locs : _np.recarray
        Localizations list.
    r : float
        Radius.
    
    Returns
    -------
    is_picked : _np.ndarray
        Boolean array - True if a localization is within radius r
        of position (x, y).
    """
    
    dx = locs.x - x
    dy = locs.y - y
    r2 = r**2
    is_picked = dx**2 + dy**2 < r2
    return is_picked


def locs_at(x: float, y: float, locs: _np.recarray, r: float) -> _np.recarray:
    """Returns localizations within radius r of position (x, y).

    Parameters
    ----------
    x : float
        x-coordinate of the position.
    y : float
        y-coordinate of the position.
    locs : _np.recarray
        Localizations list.
    r : float   
        Radius.

    Returns 
    -------
    picked_locs : _np.recarray
        Localizations in the specified area.
    """

    is_picked = is_loc_at(x, y, locs, r)
    picked_locs = locs[is_picked]
    return picked_locs


@_numba.jit(nopython=True)
def check_if_in_polygon(
    x: _np.ndarray, 
    y: _np.ndarray, 
    X: _np.ndarray, 
    Y: _np.ndarray,
) -> _np.ndarray:
    """Checks if points (x, y) are within the polygon defined by corners 
    (X, Y). Uses the ray casting algorithm, see 
    ``check_if_in_rectangle`` for details.
    
    Parameters
    ----------
    x, y : _np.ndarray
        x and y coordinates of points.
    X, Y : _np.ndarray
        x and y coordinates of polygon corners.

    Returns
    ------- 
    is_in_polygon : _np.ndarray
        Boolean array indicating which points are in the polygon.
    """

    n_locs = len(x)
    n_polygon = len(X)
    is_in_polygon = _np.zeros(n_locs, dtype=_np.bool_)

    for i in range(n_locs):
        count = 0
        for j in range(n_polygon):
            j_next = (j + 1) % n_polygon
            if (
                ((Y[j] > y[i]) != (Y[j_next] > y[i])) and
                (x[i] < X[j] + (X[j_next] - X[j]) * (y[i] - Y[j]) / (Y[j_next] - Y[j]))
            ):
                count += 1
        if count % 2 == 1:
            is_in_polygon[i] = True

    return is_in_polygon


def locs_in_polygon(
    locs: _np.recarray, 
    X: _np.ndarray, 
    Y: _np.ndarray,
) -> _np.recarray:
    """Returns localizations within the polygon defined by corners 
    (X, Y).
    
    Parameters
    ----------
    locs : _np.recarray
        Localizations.
    X : list
        x-coordinates of polygon corners.
    Y : list
        y-coordinates of polygon corners.
    
    Returns
    -------
    picked_locs : _np.recarray
        Localizations in polygon.
    """

    is_in_polygon = check_if_in_polygon(
        locs.x, locs.y, _np.array(X), _np.array(Y)
    )
    return locs[is_in_polygon]


@_numba.jit(nopython=True)
def check_if_in_rectangle(
    x: _np.ndarray, 
    y: _np.ndarray, 
    X: _np.ndarray, 
    Y: _np.ndarray,
) -> _np.ndarray:
    """Checks if locs with coordinates (x, y) are in rectangle with 
    corners (X, Y) by counting the number of rectangle sides which are 
    hit by a ray originating from each loc to the right. If the number 
    of hit rectangle sides is odd, then the loc is in the rectangle.

    Parameters
    ----------
    x, y : _np.ndarray
        x and y coordinates of points.
    X, Y : _np.ndarray
        x and y coordinates of polygon corners.

    Returns
    ------- 
    is_in_polygon : _np.ndarray
        Boolean array indicating if point is in polygon.
    """

    n_locs = len(x)
    ray_hits_rectangle_side = _np.zeros((n_locs, 4))
    for i in range(4):
        # get two y coordinates of corner points forming one rectangle side
        y_corner_1 = Y[i]
        # take the first if we're at the last side:
        y_corner_2 = Y[0] if i == 3 else Y[i + 1]
        y_corners_min = min(y_corner_1, y_corner_2)
        y_corners_max = max(y_corner_1, y_corner_2)
        for j in range(n_locs):
            y_loc = y[j]
            # only if loc is on level of rectangle side, its ray can hit:
            if y_corners_min <= y_loc <= y_corners_max:
                x_corner_1 = X[i]
                # take the first if we're at the last side:
                x_corner_2 = X[0] if i == 3 else X[i + 1]
                # calculate intersection point of ray and side:
                m_inv = (x_corner_2 - x_corner_1) / (y_corner_2 - y_corner_1)
                x_intersect = m_inv * (y_loc - y_corner_1) + x_corner_1
                x_loc = x[j]
                if x_intersect >= x_loc:
                    # ray hits rectangle side on the right side
                    ray_hits_rectangle_side[j, i] = 1
    n_sides_hit = _np.sum(ray_hits_rectangle_side, axis=1)
    is_in_rectangle = n_sides_hit % 2 == 1
    return is_in_rectangle


def locs_in_rectangle(
    locs: _np.recarray, 
    X: _np.ndarray, 
    Y: _np.ndarray,
) -> _np.recarray:
    """Returns localizations within the rectangle defined by corners 
    (X, Y).
    
    Parameters
    ----------
    locs : _np.recarray
        Localizations list.
    X, Y : list
        x and y coordinates of rectangle corners.

    Returns
    -------
    picked_locs : _np.recarray
        Localizations in rectangle.
    """
    
    is_in_rectangle = check_if_in_rectangle(
        locs.x, locs.y, _np.array(X), _np.array(Y)
    )
    picked_locs = locs[is_in_rectangle]
    return picked_locs


def minimize_shifts(
    shifts_x: _np.ndarray, 
    shifts_y: _np.ndarray, 
    shifts_z: _np.ndarray | None = None,
) -> tuple[_np.ndarray, _np.ndarray, _np.ndarray | None]:
    """Minimizes shifts in x, y, and z directions. Used for drift 
    correction.
    
    Parameters
    ---------- 
    shifts_x, shifts_y : _np.ndarray
        Shifts in x and y directions, shape (n_channels, n_channels).
    shifts_z : _np.ndarray, optional
        Shifts in z direction, shape (n_channels, n_channels). If None,
        only x and y shifts are minimized.

    Returns
    -------
    shift_y : _np.ndarray
        Minimized shifts in y direction.
    shift_x : _np.ndarray
        Minimized shifts in x direction.
    shift_z : _np.ndarray, optional
        Minimized shifts in z direction if shifts_z is specified.
    """
    
    n_channels = shifts_x.shape[0]
    n_pairs = int(n_channels * (n_channels - 1) / 2)
    n_dims = 2 if shifts_z is None else 3
    rij = _np.zeros((n_pairs, n_dims))
    A = _np.zeros((n_pairs, n_channels - 1))
    flag = 0
    for i in range(n_channels - 1):
        for j in range(i + 1, n_channels):
            rij[flag, 0] = shifts_y[i, j]
            rij[flag, 1] = shifts_x[i, j]
            if n_dims == 3:
                rij[flag, 2] = shifts_z[i, j]
            A[flag, i:j] = 1
            flag += 1
    Dj = _np.dot(_np.linalg.pinv(A), rij)
    shift_y = _np.insert(_np.cumsum(Dj[:, 0]), 0, 0)
    shift_x = _np.insert(_np.cumsum(Dj[:, 1]), 0, 0)
    if n_dims == 2:
        return shift_y, shift_x
    else:
        shift_z = _np.insert(_np.cumsum(Dj[:, 2]), 0, 0)
        return shift_y, shift_x, shift_z


def n_futures_done(futures: list[_Future]) -> int:
    """Returns the number of finished futures, used in 
    multiprocessing."""

    return sum([_.done() for _ in futures])


def remove_from_rec(rec_array: _np.recarray, name: str) -> _np.recarray:
    """Removes a column from the existing recarray.

    Parameters
    ----------
    rec_array : _np.recarray
        Recarray from which the column is removed.
    name : str
        Name of the column to be removed.
    
    Returns
    -------
    rec_array : _np.recarray
        Recarray without the column.
    """

    rec_array = _drop_fields(rec_array, name, usemask=False, asrecarray=True)
    return rec_array


def locs_glob_map(
    func: Callable[[_np.recarray, dict, str, _Any], tuple[_np.recarray, list[dict]]], 
    pattern: str, 
    args: list = [], 
    kwargs: dict = {}, 
    extension: str = "",
) -> None:
    """Maps a function to localization files, specified by a unix style 
    path pattern.
    The function must take two arguments: locs and info. It may take 
    additional args and kwargs which are supplied to this map function.
    A new locs file will be saved if an extension is provided. In that 
    case the mapped function must return new locs and a new info dict.

    Parameters
    ----------
    func : Callable
        Function to be mapped to each locs file. It must take
        locs, info, path, and any additional args and kwargs.
    pattern : str
        Unix style path pattern to match locs files.
    args : list, optional
        Additional positional arguments to be passed to the function.
    kwargs : dict, optional
        Additional keyword arguments to be passed to the function.
    extension : str, optional
        If provided, the mapped function must return new locs and info 
        dict, and a new locs file will be saved with this extension.
        If not provided, the function is expected to modify locs and 
        info in place.
    """

    paths = _glob.glob(pattern)
    for path in paths:
        locs, info = _io.load_locs(path)
        result = func(locs, info, path, *args, **kwargs)
        if extension:
            base, ext = _ospath.splitext(path)
            out_path = base + "_" + extension + ".hdf5"
            locs, info = result
            _io.save_locs(out_path, locs, info)


def get_pick_polygon_corners(
    pick: list[tuple[float, float]],
) -> tuple[list[float], list[float]]:
    """Returns X and Y coordinates of a pick polygon.
    Returns None, None if the pick is not a closed polygon.
    
    Parameters
    ----------
    pick : list of tuples
        List of tuples, each tuple contains x and y coordinates of a
        polygon corner.
    
    Returns
    -------
    X, Y : list of floats
        Lists of x and y coordinates of the polygon corners.
        Returns None, None if the pick is not a closed polygon.
    """

    if len(pick) < 3 or pick[0] != pick[-1]:
        return None, None
    else:
        X = [_[0] for _ in pick]
        Y = [_[1] for _ in pick]
        return X, Y


def get_pick_rectangle_corners(
    start_x: float, 
    start_y: float, 
    end_x: float, 
    end_y: float, 
    width: float,
) -> tuple[list[float], list[float]]:
    """Finds the positions of corners of a rectangular pick.
    Rectangular pick is defined by:
        [(start_x, start_y), (end_x, end_y)]
    and its width. (all values in camera pixels).

    Parameters
    ----------
    start_x, start_y : float
        Starting point of the pick.
    end_x, end_y : float
        Ending point of the pick.
    width : float
        Width of the pick in camera pixels.

    Returns
    -------
    corners : tuple
        Contains corners' x and y coordinates in two lists.
    """

    if end_x == start_x:
        alpha = _np.pi / 2
    else:
        alpha = _np.arctan((end_y - start_y) / (end_x - start_x))
    dx = width * _np.sin(alpha) / 2
    dy = width * _np.cos(alpha) / 2
    x1 = float(start_x - dx)
    x2 = float(start_x + dx)
    x4 = float(end_x - dx)
    x3 = float(end_x + dx)
    y1 = float(start_y + dy)
    y2 = float(start_y - dy)
    y4 = float(end_y + dy)
    y3 = float(end_y - dy)
    corners = ([x1, x2, x3, x4], [y1, y2, y3, y4])
    return corners


# def pick_areas_circle(picks, r):
#     """Returns pick areas for each pick in picks.
    
#     Parameters
#     ----------
#     picks : list
#         List of picks, each pick is a list of x and y coordinates.
#     r : float
#         Pick radius.
    
#     Returns
#     -------
#     areas : _np.ndarray
#         Pick areas, same units as r.
#     """

#     areas = _np.ones(len(picks)) * _np.pi * r**2
#     return areas


def polygon_area(X: _np.ndarray, Y: _np.ndarray) -> float:
    """Finds the area of a polygon defined by corners X and Y.
    
    Parameters
    ----------
    X, Y : _np.ndarray
        x-coordinates and y-coordinates of the polygon corners.

    Returns
    -------
    area : float
        Area of the polygon.
    """

    n_corners = len(X)
    area = 0
    for i in range(n_corners):
        j = (i + 1) % n_corners # next corner
        area += X[i] * Y[j] - X[j] * Y[i]
    area = abs(area) / 2
    return area


def pick_areas_polygon(picks: list[list[tuple[float, float]]]) -> _np.ndarray:
    """Returns pick areas for each polygonal pick in picks.
    
    Parameters
    ----------
    picks : list of lists of tuples
        List of picks, each pick is a list of (x, y) coordinates of the 
        polygon corners.
    
    Returns
    -------
    areas : _np.ndarray
        Pick areas.
    """

    areas = []
    for i, pick in enumerate(picks):
        if len(pick) < 3 or pick[0] != pick[-1]: # not a closed polygon
            continue
        X, Y = get_pick_polygon_corners(pick)
        areas.append(polygon_area(X, Y))
    areas = _np.array(areas)
    areas = areas[areas > 0] # remove open polygons 
    return areas


def pick_areas_rectangle(
    picks: list[list[tuple[float, float]]], 
    w: float,
) -> _np.ndarray:
    """Returns pick areas for each pick in picks.
    
    Parameters
    ----------
    picks : list
        List of picks, each pick is a list of coordinates of the 
        rectangle corners.
    w : float
        Pick width.
    
    Returns
    -------
    areas : _np.ndarray
        Pick areas, same units as w.
    """

    areas = _np.zeros(len(picks))
    for i, pick in enumerate(picks):
        (xs, ys), (xe, ye) = pick
        areas[i] = w * _np.sqrt((xe - xs) ** 2 + (ye - ys) ** 2)
    return areas


"""
    picasso.gausslq
    ~~~~~~~~~~~~~~~~

    Fit spots with Gaussian least squares

    :authors: Joerg Schnitzbauer, Maximilian Thomas Strauss, 2016-2018
    :copyright: Copyright (c) 2016-2018 Jungmann Lab, MPI of Biochemistry
"""


from scipy import optimize as _optimize
import numpy as _np
from tqdm import tqdm as _tqdm
import numba as _numba
import multiprocessing as _multiprocessing
from concurrent import futures as _futures
from . import postprocess as _postprocess

try:
    from pygpufit import gpufit as gf

    gpufit_installed = True
except ImportError:
    gpufit_installed = False


@_numba.jit(nopython=True, nogil=True)
def _gaussian(mu: float, sigma: float, grid: _np.ndarray) -> _np.ndarray:
    """Compute a Gaussian PDF on a grid."""

    norm = 0.3989422804014327 / sigma
    return norm * _np.exp(-0.5 * ((grid - mu) / sigma) ** 2)


"""
def integrated_gaussian(mu, sigma, grid):
    norm = 0.70710678118654757 / sigma   # sq_norm = sqrt(0.5/sigma**2)
    integrated_gaussian =  0.5 *
    (erf((grid - mu + 0.5) * norm) - erf((grid - mu - 0.5) * norm))
    return integrated_gaussian
"""


@_numba.jit(nopython=True, nogil=True)
def _sum_and_center_of_mass(
    spot: _np.ndarray, 
    size: int,
) -> tuple[float, float, float]:
    """Calculate the sum and center of mass of a 2D spot."""

    x = 0.0
    y = 0.0
    _sum_ = 0.0
    for i in range(size):
        for j in range(size):
            x += spot[i, j] * i
            y += spot[i, j] * j
            _sum_ += spot[i, j]
    x /= _sum_
    y /= _sum_
    return _sum_, y, x


@_numba.jit(nopython=True, nogil=True)
def _initial_sigmas(
    spot: _np.ndarray, 
    y: float, 
    x: float, 
    sum: float, 
    size: int,
) -> tuple[float, float]:
    """Initialize the sizes of the single-emitter images (sigmas of the
    Gaussian fit) in x and y independently."""

    sum_deviation_y = 0.0
    sum_deviation_x = 0.0
    for i in range(size):
        for j in range(size):
            sum_deviation_y += spot[i, j] * (i - y) ** 2
            sum_deviation_x += spot[i, j] * (j - x) ** 2
    sy = _np.sqrt(sum_deviation_y / sum)
    sx = _np.sqrt(sum_deviation_x / sum)
    return sy, sx


@_numba.jit(nopython=True, nogil=True)
def _initial_parameters(
    spot: _np.ndarray, 
    size: int, 
    size_half: int,
) -> _np.ndarray:
    """Initialize the parameters for the Gaussian fit - x, y, photons, 
    background, sigma_x, sigma_y."""

    theta = _np.zeros(6, dtype=_np.float32)
    theta[3] = _np.min(spot)
    spot_without_bg = spot - theta[3]
    sum, theta[1], theta[0] = _sum_and_center_of_mass(spot_without_bg, size)
    theta[2] = _np.maximum(1.0, sum)
    theta[5], theta[4] = _initial_sigmas(
        spot - theta[3], theta[1], theta[0], sum, size
    )
    theta[0:2] -= size_half
    return theta


def initial_parameters_gpufit(spots: _np.ndarray, size: int) -> _np.ndarray:
    """Initialize the parameters for the GPU fit - photons, x, y, sx, 
    sy, bg."""

    center = (size / 2.0) - 0.5
    initial_width = _np.amax([size / 5.0, 1.0])

    spot_max = _np.amax(spots, axis=(1, 2))
    spot_min = _np.amin(spots, axis=(1, 2))

    initial_parameters = _np.empty((len(spots), 6), dtype=_np.float32)

    initial_parameters[:, 0] = spot_max - spot_min
    initial_parameters[:, 1] = center
    initial_parameters[:, 2] = center
    initial_parameters[:, 3] = initial_width
    initial_parameters[:, 4] = initial_width
    initial_parameters[:, 5] = spot_min

    return initial_parameters


@_numba.jit(nopython=True, nogil=True)
def _outer(
    a: _np.ndarray, 
    b: _np.ndarray, 
    size: int, 
    model: _np.ndarray, 
    n: float, 
    bg: float
) -> None:
    """Compute the outer product of two vectors a and b, scaled by n and
    added a background value bg, and store the result in model."""

    for i in range(size):
        for j in range(size):
            model[i, j] = n * a[i] * b[j] + bg


@_numba.jit(nopython=True, nogil=True)
def _compute_model(
    theta: _np.ndarray, 
    grid: _np.ndarray, 
    size: int, 
    model_x: _np.ndarray, 
    model_y: _np.ndarray, 
    model: _np.ndarray,
) -> _np.ndarray:
    """Compute the model of a Gaussian spot (2D) based on the parameters 
    in theta, which contains the x and y positions, the number of 
    photons, background, and the sigmas in x and y."""

    model_x[:] = _gaussian(
        theta[0], theta[4], grid
    )  # sx and sy are wrong with integrated gaussian
    model_y[:] = _gaussian(
        theta[1], theta[5], grid
    )
    _outer(model_y, model_x, size, model, theta[2], theta[3])
    return model


@_numba.jit(nopython=True, nogil=True)
def _compute_residuals(
    theta: _np.ndarray, 
    spot: _np.ndarray, 
    grid: _np.ndarray, 
    size: int, 
    model_x: _np.ndarray, 
    model_y: _np.ndarray, 
    model: _np.ndarray, 
    residuals: _np.ndarray,
) -> _np.ndarray:
    """Compute the residuals (i.e., the difference in pixel values) 
    between the observed spot and the model computed from the parameters 
    in theta."""

    _compute_model(theta, grid, size, model_x, model_y, model)
    residuals[:, :] = spot - model
    return residuals.flatten()


def fit_spot(spot: _np.ndarray) -> _np.ndarray:
    """Fit a single spot using least squares optimization. The spot is a
    2D array representing the pixel values of the spot image. The 
    function returns the optimized parameters as a 1D array with the
    following order: [x, y, photons, bg, sx, sy].
    
    The parameters are initialized based on the spot's pixel values, and
    the optimization is performed using the least squares method. The
    optimization minimizes the residuals between the observed spot and
    the model computed from the parameters.
    
    Parameters
    ----------
    spot : _np.ndarray
        A 2D array representing the pixel values of the spot image. 
        The shape of the array should be (size, size), where size is the
        length of one side of the square spot image.

    Returns
    -------
    result_ : _np.ndarray
        A 1D array containing the optimized parameters in the following order:
        [x, y, photons, bg, sx, sy].
    """

    size = spot.shape[0]
    size_half = int(size / 2)
    grid = _np.arange(-size_half, size_half + 1, dtype=_np.float32)
    model_x = _np.empty(size, dtype=_np.float32)
    model_y = _np.empty(size, dtype=_np.float32)
    model = _np.empty((size, size), dtype=_np.float32)
    residuals = _np.empty((size, size), dtype=_np.float32)
    # theta is [x, y, photons, bg, sx, sy]
    theta0 = _initial_parameters(spot, size, size_half)
    args = (spot, grid, size, model_x, model_y, model, residuals)
    result = _optimize.leastsq(
        _compute_residuals, theta0, args=args, ftol=1e-2, xtol=1e-2
    )  # leastsq is much faster than least_squares
    """
    model = compute_model(result[0], grid, size, model_x, model_y, model)
    plt.figure()
    plt.subplot(121)
    plt.imshow(spot, interpolation='none')
    plt.subplot(122)
    plt.imshow(model, interpolation='none')
    plt.colorbar()
    plt.show()
    """
    result_ = result[0]
    return result_


def fit_spots(spots: _np.ndarray) -> _np.ndarray:
    """Fit multiple spots using least squares optimization. Each spot is
    a 2D array representing the pixel values of the spot image. The
    function returns a 2D array with the optimized parameters for each
    spot, where each row corresponds to a spot and the columns are the
    parameters in the following order: [x, y, photons, bg, sx, sy].

    Parameters
    ----------
    spots : _np.ndarray
        A 3D array of shape (n_spots, size, size), where n_spots is the
        number of spots and size is the length of one side of the square
        spot image. Each slice along the first axis represents a single
        spot image.
    
    Returns
    -------
    theta : _np.ndarray
        A 2D array with the optimized parameters for each spot. The
        columns correspond to [x, y, photons, bg, sx, sy].
    """

    theta = _np.empty((len(spots), 6), dtype=_np.float32)
    theta.fill(_np.nan)
    for i, spot in enumerate(spots):
        theta[i] = fit_spot(spot)
    return theta


def fit_spots_parallel(
    spots: _np.ndarray, 
    asynch: bool = False,
) -> _np.ndarray | list[_futures.Future]:
    """Allows for running ``fit_spots`` asynchronously 
    (multiprocessing).
    
    Parameters
    ----------
    spots : _np.ndarray
        A 3D array of shape (n_spots, size, size), where n_spots is the
        number of spots and size is the length of one side of the square
        spot image. Each slice along the first axis represents a single
        spot image.
    asynch : bool, optional
        If True, the function returns a list of futures that can be
        processed asynchronously. If False, the function waits for all
        futures to complete and returns the results as a 2D array.
        
    Returns
    -------
    _np.ndarray | list[_futures.Future]
        If `asynch` is False, returns a 2D array with the optimized
        parameters for each spot, where each row corresponds to a spot
        and the columns are the parameters in the following order:
        [x, y, photons, bg, sx, sy]. If `asynch` is True, returns a list 
        of futures that can be processed asynchronously.
    """

    n_workers = min(
        60, max(1, int(0.75 * _multiprocessing.cpu_count()))
    ) # Python crashes when using >64 cores
    n_spots = len(spots)
    n_tasks = 100 * n_workers
    spots_per_task = [
        int(n_spots / n_tasks + 1) if _ < n_spots % n_tasks else int(n_spots / n_tasks)
        for _ in range(n_tasks)
    ]
    start_indices = _np.cumsum([0] + spots_per_task[:-1])
    fs = []
    executor = _futures.ProcessPoolExecutor(n_workers)
    for i, n_spots_task in zip(start_indices, spots_per_task):
        fs.append(executor.submit(fit_spots, spots[i : i + n_spots_task]))
    if asynch:
        return fs
    with _tqdm(desc="LQ fitting", total=n_tasks, unit="task") as progress_bar:
        for f in _futures.as_completed(fs):
            progress_bar.update()
    return fits_from_futures(fs)


def fit_spots_gpufit(spots: _np.ndarray) -> _np.ndarray:
    """Fit multiple spots using GPU-based Gaussian fitting. Each spot is
    a 2D array representing the pixel values of the spot image. The
    function returns a 2D array with the optimized parameters for each
    spot, where each row corresponds to a spot and the columns are the
    parameters in the following order: [photons, x, y, sx, sy, bg].

    Cite: Przybylski, et al. Scientific Reports, 2017. 
    DOI: 10.1038/s41598-017-15313-9
    
    Parameters
    ----------
    spots : _np.ndarray
        A 3D array of shape (n_spots, size, size), where n_spots is the
        number of spots and size is the length of one side of the square
        spot image. Each slice along the first axis represents a single
        spot image.
        
    Returns
    -------
    parameters : _np.ndarray
        A 2D array with the optimized parameters for each spot. The
        columns correspond to [photons, x, y, sx, sy, bg].
    """

    size = spots.shape[1]
    initial_parameters = initial_parameters_gpufit(spots, size)
    spots.shape = (len(spots), (size * size))
    model_id = gf.ModelID.GAUSS_2D_ELLIPTIC

    parameters, states, chi_squares, number_iterations, exec_time = gf.fit(
        spots,
        None,
        model_id,
        initial_parameters,
        tolerance=1e-2,
        max_number_iterations=20,
    )

    parameters[:, 0] *= 2.0 * _np.pi * parameters[:, 3] * parameters[:, 4]

    return parameters


def fits_from_futures(futures: list[_futures.Future]) -> _np.ndarray:
    """Collect results from futures and stack them into a 2D array."""
    
    theta = [_.result() for _ in futures]
    return _np.vstack(theta)


def locs_from_fits(
    identifications: _np.recarray, 
    theta: _np.ndarray, 
    box: int, 
    em: bool,
) -> _np.recarray:
    """Convert the fit results into a structured array of localizations.
    
    Parameters
    ----------
    identifications : _np.recarray
        A structured array containing the identifications of the spots,
        including frame numbers, x and y coordinates, and net gradient.
    theta : _np.ndarray
        A 2D array with the optimized parameters for each spot, where
        each row corresponds to a spot and the columns are the 
        parameters in the following order: [x, y, photons, bg, sx, sy].
    box : int   
        The size of the box used for localization, which is used to
        calculate the offsets for the x and y coordinates.
    em : bool
        Whether EMCCD was used for the localization.
    
    Returns
    -------
    locs : _np.recarray
        A structured array containing the localized spots.
    """

    # box_offset = int(box/2)
    x = theta[:, 0] + identifications.x  # - box_offset
    y = theta[:, 1] + identifications.y  # - box_offset
    lpx = _postprocess.localization_precision(
        theta[:, 2], theta[:, 4], theta[:, 3], em=em
    )
    lpy = _postprocess.localization_precision(
        theta[:, 2], theta[:, 5], theta[:, 3], em=em
    )
    a = _np.maximum(theta[:, 4], theta[:, 5])
    b = _np.minimum(theta[:, 4], theta[:, 5])
    ellipticity = (a - b) / a

    if hasattr(identifications, "n_id"):
        locs = _np.rec.array(
            (
                identifications.frame,
                x,
                y,
                theta[:, 2],
                theta[:, 4],
                theta[:, 5],
                theta[:, 3],
                lpx,
                lpy,
                ellipticity,
                identifications.net_gradient,
                identifications.n_id,
            ),
            dtype=[
                ("frame", "u4"),
                ("x", "f4"),
                ("y", "f4"),
                ("photons", "f4"),
                ("sx", "f4"),
                ("sy", "f4"),
                ("bg", "f4"),
                ("lpx", "f4"),
                ("lpy", "f4"),
                ("ellipticity", "f4"),
                ("net_gradient", "f4"),
                ("n_id", "u4"),
            ],
        )
        locs.sort(kind="mergesort", order="n_id")
    else:
        locs = _np.rec.array(
            (
                identifications.frame,
                x,
                y,
                theta[:, 2],
                theta[:, 4],
                theta[:, 5],
                theta[:, 3],
                lpx,
                lpy,
                ellipticity,
                identifications.net_gradient,
            ),
            dtype=[
                ("frame", "u4"),
                ("x", "f4"),
                ("y", "f4"),
                ("photons", "f4"),
                ("sx", "f4"),
                ("sy", "f4"),
                ("bg", "f4"),
                ("lpx", "f4"),
                ("lpy", "f4"),
                ("ellipticity", "f4"),
                ("net_gradient", "f4"),
            ],
        )
        locs.sort(kind="mergesort", order="frame")
    return locs


def locs_from_fits_gpufit(
    identifications: _np.recarray, 
    theta: _np.ndarray, 
    box: int, 
    em: bool,
) -> _np.recarray:
    """Convert the fit results from GPU-based fitting into a structured
    array of localizations.
    
    Parameters
    ----------
    identifications : _np.recarray
        A structured array containing the identifications of the spots,
        including frame numbers, x and y coordinates, and net gradient.
    theta : _np.ndarray
        A 2D array with the optimized parameters for each spot, where
        each row corresponds to a spot and the columns are the 
        parameters in the following order: [photons, x, y, sx, sy, bg].
    box : int
        The size of the box used for localization, which is used to
        calculate the offsets for the x and y coordinates.
    em : bool
        Whether EMCCD was used for the localization.   
    
    Returns
    -------
    locs : _np.recarray
        A structured array containing the localized spots.
    """
    
    box_offset = int(box / 2)
    x = theta[:, 1] + identifications.x - box_offset
    y = theta[:, 2] + identifications.y - box_offset
    lpx = _postprocess.localization_precision(
        theta[:, 0], theta[:, 3], theta[:, 5], em=em
    )
    lpy = _postprocess.localization_precision(
        theta[:, 0], theta[:, 4], theta[:, 5], em=em
    )
    a = _np.maximum(theta[:, 3], theta[:, 4])
    b = _np.minimum(theta[:, 3], theta[:, 4])
    ellipticity = (a - b) / a
    locs = _np.rec.array(
        (
            identifications.frame,
            x,
            y,
            theta[:, 0],
            theta[:, 3],
            theta[:, 4],
            theta[:, 5],
            lpx,
            lpy,
            ellipticity,
            identifications.net_gradient,
        ),
        dtype=[
            ("frame", "u4"),
            ("x", "f4"),
            ("y", "f4"),
            ("photons", "f4"),
            ("sx", "f4"),
            ("sy", "f4"),
            ("bg", "f4"),
            ("lpx", "f4"),
            ("lpy", "f4"),
            ("ellipticity", "f4"),
            ("net_gradient", "f4"),
        ],
    )
    locs.sort(kind="mergesort", order="frame")
    return locs



"""
    picasso.clusterer
    ~~~~~~~~~~~~~~~~~

    Clusterer optimized for DNA PAINT, as well as DBSCAN and HDBSCAN.

    Based on the work of Thomas Schlichthaerle and Susanne Reinhardt.
    :authors: Thomas Schlichthaerle, Susanne Reinhardt, 
        Rafal Kowalewski, 2020-2022
    :copyright: Copyright (c) 2022 Jungmann Lab, MPI of Biochemistry
"""

from __future__ import annotations

import numpy as _np
import pandas as _pd
from scipy.spatial import cKDTree as _cKDTree
from scipy.spatial import ConvexHull as _ConvexHull
from sklearn.cluster import DBSCAN as _DBSCAN
from sklearn.cluster import HDBSCAN as _HDBSCAN

from . import lib as _lib

CLUSTER_CENTERS_DTYPE_2D = [
    ("frame", "f4"),
    ("std_frame", "f4"),
    ("x", "f4"),
    ("y", "f4"),
    ("std_x", "f4"),
    ("std_y", "f4"),
    ("photons", "f4"),
    ("sx", "f4"),
    ("sy", "f4"),
    ("bg", "f4"),
    ("lpx", "f4"),
    ("lpy", "f4"),
    ("ellipticity", "f4"),
    ("net_gradient", "f4"),
    ("n", "u4"),
    ("n_events", "i4"), 
    ("area", "f4"),
    ("convexhull", "f4"),
    ("group", "i4"),
] 
CLUSTER_CENTERS_DTYPE_3D = [
    ("frame", "f4"),
    ("std_frame", "f4"),
    ("x", "f4"),
    ("y", "f4"),
    ("std_x", "f4"),
    ("std_y", "f4"),
    ("z", "f4"),
    ("photons", "f4"),
    ("sx", "f4"),
    ("sy", "f4"),
    ("bg", "f4"),
    ("lpx", "f4"),
    ("lpy", "f4"),
    ("std_z", "f4"),
    ("ellipticity", "f4"),
    ("net_gradient", "f4"),
    ("n", "u4"),
    ("n_events", "u4"), 
    ("volume", "f4"),
    ("convexhull", "f4"),
    ("group", "i4"),
]

def _frame_analysis(frame: _pd.SeriesGroupBy, n_frames: int) -> int:
    """Verifies which clusters pass basic frame analysis.

    Rejects clusters whose mean frame is outside of the 
    [20, 80] % (max frame) range or any 1/20th of measurement's time
    contains more than 80 % of localizations.

    Assumes frame to be a pandas.SeriesGroupBy object, grouped by
    cluster ids.

    Parameters
    ----------
    frame : pandas.SeriesGroupBy
        Frame number for a given cluster; grouped by cluster ids.
    n_frames : int
        Acquisition time given in frames.

    Returns
    -------
    passed : int
        1 if passed frame analysis, 0 otherwise.
    """

    passed = 1

    # get mean frame
    mean_frame = frame.values.mean()

    # get maximum number of locs in a 1/20th of acquisition time
    n_locs = len(frame)
    locs_binned = _np.histogram(
        frame.values, bins=_np.linspace(0, n_frames, 21)
    )[0]
    max_locs_bin = locs_binned.max()

    # test if frame analysis passed
    if (
        (mean_frame < 0.2 * n_frames)
        or (mean_frame > 0.8 * n_frames)
        or (max_locs_bin > 0.8 * n_locs)
    ):
        passed = 0

    return passed


def frame_analysis(labels: _np.ndarray, frame: _np.ndarray) -> _np.ndarray:
    """Performs basic frame analysis on clustered localizations.

    Rejects clusters whose mean frame is outside of the 
    [20, 80] % (max frame) range or any 1/20th of measurement's time
    contains more than 80 % of localizations.

    Uses pandas for fast calculations using groupby().

    Parameters
    ----------
    labels : _np.ndarray
        Cluster labels (-1 means no cluster assigned).
    frame : _np.ndarray
        Frame number for each localization.

    Returns
    -------
    labels : _np.ndarray
        Cluster labels for each localization (-1 means no cluster 
        assigned).
    """

    # group frames by cluster ids
    frame_pd = _pd.Series(frame, index=labels)
    frame_grouped = frame_pd.groupby(frame_pd.index)

    # perform frame analysis
    true_cluster = frame_grouped.apply(_frame_analysis, frame.max()+1)

    # cluster ids that did not pass frame analysis
    discard = true_cluster.index[true_cluster == 0].values
    # change labels of these clusters to -1
    labels[_np.isin(labels, discard)] = -1

    return labels


def _cluster(
    X: _np.ndarray, 
    radius: float, 
    min_locs: int, 
    frame: _np.ndarray | None = None,
) -> _np.ndarray:
    """Clusters points given by X with a given clustering radius and
    minimum number of localizations within that radius using KDTree.

    The general workflow is as follows:
    1. Build a KDTree from the points in X.
    2. For each point, find its neighbors within the given radius.
    3. Identify local maxima, i.e., points with the most neighbors
       within their neighborhood.
    4. Assign cluster labels to all points. If two local maxima are
       within the radius from each other, combine such clusters.

    Based on the algorithm published by Schichthaerle et al.
    Nature Comm, 2021 (10.1038/s41467-021-22606-1) and first implemented
    in Reinhardt, Masullo, Baudrexel, Steen, et al. Nature, 2023 
    (10.1038/s41586-023-05925-9).

    Parameters
    ----------
    X : _np.ndarray
        Array of points of shape (n_points, n_dim) to be clustered.
    radius : float
        Clustering radius.
    min_locs : int
        Minimum number of localizations in a cluster.
    frame : _np.ndarray, optional
        Frame number of each localization. If None, no frame analysis
        is performed.

    Returns
    -------
    labels : _np.ndarray
        Cluster labels for each localization (-1 means no cluster 
        assigned).
    """

    ## build kdtree (use cKDTree in case user did not update scipy)
    tree = _cKDTree(X)

    ## find neighbors for each point within radius
    neighbors = tree.query_ball_tree(tree, radius)

    ## find local maxima, i.e., points with the most neighbors within
    ## their neighborhood
    lm = _np.zeros(X.shape[0], dtype=_np.int8)
    for i in range(len(lm)):
        idx = neighbors[i] # indeces of points that are neighbors of i
        n = len(idx) # number of neighbors of i
        if n > min_locs: # note that i is included in its neighbors
            # if i has the most neighbors in its neighborhood
            if n == max([len(neighbors[_]) for _ in idx]):
                lm[i] = 1

    ## assign cluster labels to all points (-1 means no cluster)
    ## if two local maxima are within radius from each other, combine
    ## such clusters
    labels = -1 * _np.ones(X.shape[0], dtype=_np.int32) # cluster labels
    lm_idx = _np.where(lm == 1)[0] # indeces of local maxima

    for count, i in enumerate(lm_idx): # for each local maximum
        label = labels[i]
        if label == -1: # if lm not assigned yet
            labels[neighbors[i]] = count
        else:
            # indeces of locs that were not assigned to any cluster
            idx = [
                neighbors[i][_]
                for _ in _np.where(labels[neighbors[i]] == -1)[0]
            ]
            if len(idx): # if such a loc exists, assign it to a cluster
                labels[idx] = label

    ## check for number of locs per cluster to be above min_locs
    values, counts = _np.unique(labels, return_counts=True)
    # labels to discard if has fewer locs than min_locs
    to_discard = values[counts < min_locs]
    # substitute this with -1
    labels[_np.isin(labels, to_discard)] = -1

    if frame is not None:
        labels = frame_analysis(labels, frame)

    return labels


def cluster_2D(
    x: _np.ndarray, 
    y: _np.ndarray, 
    frame: _np.ndarray, 
    radius: float, 
    min_locs: int, 
    fa: bool,
) -> _np.ndarray:
    """Prepares 2D input to be used by _cluster()

    Parameters
    ----------
    x, y : _np.ndarray
        x and y coordinates to be clustered.
    frame : _np.ndarray
        Frame number for each localization.
    radius : float
        Clustering radius.
    min_locs : int
        Minimum number of localizations in a cluster.
    fa : bool
        True, if basic frame analysis is to be performed.

    Returns
    -------
    labels : _np.ndarray
        Cluster labels for each localization (-1 means no cluster 
        assigned).
    """

    X = _np.stack((x, y)).T

    if not fa:
        frame = None

    labels = _cluster(X, radius, min_locs, frame)

    return labels


def cluster_3D(x, y, z, frame, radius_xy, radius_z, min_locs, fa):
    """Prepares 3D input to be used by _cluster()

    Scales z coordinates by radius_xy / radius_z

    Parameters
    ----------
    x, y, z : _np.ndarray
        x, y, and z coordinates to be clustered.
    frame : _np.ndarray
        Frame number for each localization.
    radius_xy : float
        Clustering radius in x and y directions.
    radius_z : float
        Clustering radius in z direction.
    min_locs : int
        Minimum number of localizations in a cluster.
    fa : bool
        True, if basic frame analysis is to be performed.

    Returns
    -------
    labels : _np.ndarray
        Cluster labels for each localization (-1 means no cluster 
        assigned).
    """

    radius = radius_xy
    X = _np.stack((x, y, z * radius_xy / radius_z)).T

    if not fa:
        frame = None

    labels = _cluster(X, radius, min_locs, frame)
    
    return labels


def cluster(
    locs: _np.recarray, 
    radius_xy: float, 
    min_locs: int, 
    frame_analysis: bool, 
    radius_z: float | None = None, 
    pixelsize: float | None = None,
) -> _np.recarray:
    """Clusters localizations (SMLM clusterer).

    The general workflow is as follows:
    1. Build a KDTree from the points in X.
    2. For each point, find its neighbors within the given radius.
    3. Identify local maxima, i.e., points with the most neighbors
       within their neighborhood.
    4. Assign cluster labels to all points. If two local maxima are
       within the radius from each other, combine such clusters.

    Based on the algorithm published by Schichthaerle et al.
    Nature Comm, 2021 (10.1038/s41467-021-22606-1) and first implemented
    in Reinhardt, Masullo, Baudrexel, Steen, et al. Nature, 2023 
    (10.1038/s41586-023-05925-9).

    The recommended parameters are ``radius`` of 2*NeNA and ``min_locs`` 
    of 10. Keep in mind that the parameters may vary between 
    applications, so we encourage you to experiment with them when 
    needed. Especially ``min_locs`` may vary between datasets.  

    Parameters
    ----------
    locs : _np.recarray
        Localizations to be clustered.
    radius_xy : float
        Clustering radius in xy plane (camera pixels).   
    min_locs : int
        Minimum number of localizations in a cluster.   
    frame_analysis : bool
        If True, performs basic frame analysis.
    radius_z : float, optional
        Clustering radius in z plane (camera pixels). Only used for
        3D clustering. 
    pixelsize : int, optional
        Camera pixel size in nm. Only needed for 3D clustering.

    Returns
    -------
    locs : _np.recarray
        Clusterered localizations, with column 'group' added, which
        specifies cluster label for each localization. Noise (label -1)
        is removed.  
    """

    if hasattr(locs, "z"): # 3D
        if pixelsize is None or radius_z is None:
            raise ValueError(
                "Camera pixel size and clustering radius in z must be specified"
                " for 3D clustering."
            )
        labels = cluster_3D(
            locs.x,
            locs.y,
            locs.z / pixelsize, # convert z coordinates from nm to px
            locs.frame,
            radius_xy,
            radius_z,
            min_locs,
            frame_analysis,
        )
    else:
        labels = cluster_2D(
            locs.x,
            locs.y,
            locs.frame,
            radius_xy,
            min_locs,
            frame_analysis,
        )
    locs = extract_valid_labels(locs, labels)
    return locs


def _dbscan(X: _np.ndarray, radius: float, min_density: int) -> _np.ndarray:
    """ Finds DBSCAN cluster labels, given data points and parameters.
    
    Parameters
    ----------
    X : _np.ndarray
        Array of shape (N, D), with N being the number of data points
        and D the number of dimensions.
    radius : float
        DBSCAN search radius, often referred to as "epsilon".
    min_density : int
        Number of points within radius to consider a given point a core
        sample.

    Returns
    -------
    labels : _np.ndarray
        Cluster labels for each point. Shape: (N,). -1 means no cluster
        assigned.
    """

    db = _DBSCAN(eps=radius, min_samples=min_density).fit(X)
    return db.labels_.astype(_np.int32)


def dbscan(
    locs: _np.recarray, 
    radius: float, 
    min_samples: int, 
    pixelsize: int | None = None,
) -> _np.recarray:
    """Performs DBSCAN on localizations.

    Parameters
    ---------
    locs : _np.recarray
        Localizations to be clustered.
    radius : float
        DBSCAN search radius, often referred to as "epsilon". Same units
        as locs.
    min_samples : int
        Number of localizations within radius to consider a given point 
        a core sample.
    pixelsize : int, optional
        Camera pixel size in nm. Only needed for 3D.
    
    Returns
    -------
    locs : _np.recarray
        Clusterered localizations, with column 'group' added, which
        specifies cluster label for each localization. Noise (label -1)
        is removed.  
    """
    
    if hasattr(locs, "z"):
        if pixelsize is None:
            raise ValueError(
                "Camera pixel size must be specified as an integer for 3D"
                " clustering."
            )
        X = _np.vstack((locs.x, locs.y, locs.z / pixelsize)).T
    else:
        X = _np.vstack((locs.x, locs.y)).T
    labels = _dbscan(X, radius, min_samples)
    locs = extract_valid_labels(locs, labels)
    return locs


def _hdbscan(
    X: _np.ndarray, 
    min_cluster_size: int, 
    min_samples: int, 
    cluster_eps: float = 0,
) -> _np.ndarray:
    """Finds HDBSCAN cluster labels, given data points and parameters.

    Parameters
    ----------
    X : _np.ndarray
        Array of shape (N, D), with N being the number of data points
        and D the number of dimensions.
    min_cluster_size : int
        Minimum number of points in cluster.
    min_samples : int
        Number of points within radius to consider a given point a core
        sample.
    cluster_eps : float, optional
        Distance threshold. Clusters below this value will be merged.
    
    Returns
    -------
    labels : _np.ndarray
        Cluster labels for each point. Shape: (N,). -1 means no cluster
        assigned.
    """
    
    hdb = _HDBSCAN(
        min_samples=min_samples, 
        min_cluster_size=min_cluster_size,
        cluster_selection_epsilon=cluster_eps,
    ).fit(X)
    return hdb.labels_.astype(_np.int32)


def hdbscan(
    locs: _np.recarray, 
    min_cluster_size: int, 
    min_samples: int, 
    pixelsize: int | None = None, 
    cluster_eps: float = 0.0,
) -> _np.recarray:
    """Performs HDBSCAN on localizations.

    Parameters
    ----------
    locs : _np.recarray
        Localizations to be clustered.
    min_cluster_size : int
        Minimum number of localizations in cluster.
    min_samples : int
        Number of localizations within radius to consider a given point
        a core sample.
    pixelsize : int, optional
        Camera pixel size in nm. Only needed for 3D.
    cluster_eps : float, optional
        Distance threshold. Clusters below this value will be merged.

    Returns
    -------
    locs : _np.recarray
        Clusterered localizations, with column 'group' added, which
        specifies cluster label for each localization. Noise (label -1)
        is removed.  
    """

    if hasattr(locs, "z"):
        if pixelsize is None:
            raise ValueError(
                "Camera pixel size (nm) must be specified as an integer for 3D"
                " clustering."
            )
        X = _np.vstack((locs.x, locs.y, locs.z / pixelsize)).T
    else:
        X = _np.vstack((locs.x, locs.y)).T
    labels = _hdbscan(
        X, min_cluster_size, min_samples, cluster_eps=cluster_eps
    )
    locs = extract_valid_labels(locs, labels)
    return locs


def extract_valid_labels(
    locs: _np.recarray, 
    labels: _np.ndarray,
) -> _np.recarray:
    """Extracts localizations based on clustering results. Localizations 
    that were not clustered are excluded.

    Parameters
    ----------
    locs : _np.recarray
        Localizations to be filtered.
    labels : _np.ndarray
        Array of cluster labels for each localization. -1 means no
        cluster assignment.

    Returns
    -------
    locs : _np.recarray
        Localization list with "group" column appended, providing
        cluster label. 
    """
    
    # add cluster id to locs, as "group"
    locs = _lib.append_to_rec(locs, labels, "group")

    # -1 means no cluster assigned to a loc
    locs = locs[locs.group != -1]
    return locs


def error_sums_wtd(x: float, w: float) -> float:
    """Function used for finding localization precision for cluster
    centers.

    Parameters
    ----------
    x : float
        x or y coordinate of the cluster center.
    w : float
        weight (inverse localization precision squared).

    Returns
    -------
    float
        Weighted localization precision of the cluster center.
    """

    return (w * (x - (w * x).sum() / w.sum())**2).sum() / w.sum()


def find_cluster_centers(
    locs: _np.recarray, 
    pixelsize: float | None = None,
) -> _np.recarray:
    """Calculates cluster centers. 

    Uses pandas.groupby to quickly run across all cluster ids.

    Parameters
    ----------
    locs : _np.recarray
        Clustered localizations (contain group info)
    pixelsize : int, optional
        Camera pixel size (used for finding volume and 3D convex hull).
        Only required for 3D localizations.

    Returns
    -------
    centers : _np.recarray
        Cluster centers saved in the format of localizations.
    """

    # group locs by their cluster id (group)
    locs_pd = _pd.DataFrame(locs)
    grouplocs = locs_pd.groupby(locs_pd.group)

    # get cluster centers
    res = grouplocs.apply(cluster_center, pixelsize)
    centers_ = res.values

    # convert to recarray and save
    frame = _np.ndarray([_[0] for _ in centers_])
    std_frame = _np.ndarray([_[1] for _ in centers_])
    x = _np.ndarray([_[2] for _ in centers_])
    y = _np.ndarray([_[3] for _ in centers_])
    std_x = _np.ndarray([_[4] for _ in centers_])
    std_y = _np.ndarray([_[5] for _ in centers_])
    photons = _np.ndarray([_[6] for _ in centers_])
    sx = _np.ndarray([_[7] for _ in centers_])
    sy = _np.ndarray([_[8] for _ in centers_])
    bg = _np.ndarray([_[9] for _ in centers_])
    lpx = _np.ndarray([_[10] for _ in centers_])
    lpy = _np.ndarray([_[11] for _ in centers_])
    ellipticity = _np.ndarray([_[12] for _ in centers_])
    net_gradient = _np.ndarray([_[13] for _ in centers_])
    n = _np.ndarray([_[14] for _ in centers_])
    n_events = _np.ndarray([_[15] for _ in centers_]) # number of localizations in cluster

    if hasattr(locs, "z"):
        z = _np.ndarray([_[16] for _ in centers_])
        std_z = _np.ndarray([_[17] for _ in centers_])
        volume = _np.ndarray([_[18] for _ in centers_])
        convexhull = _np.ndarray([_[19] for _ in centers_])
        centers = _np.rec.ndarray(
            (
                frame,
                std_frame,
                x,
                y,
                std_x,
                std_y,
                z,
                photons,
                sx,
                sy,
                bg,
                lpx,
                lpy,
                std_z,
                ellipticity,
                net_gradient,
                n,
                n_events,
                volume,
                convexhull,
                res.index.values, # group id
            ),
            dtype=CLUSTER_CENTERS_DTYPE_3D,
        )
    else:
        area = _np.ndarray([_[16] for _ in centers_])
        convexhull = _np.ndarray([_[17] for _ in centers_])
        centers = _np.rec.ndarray(
            (
                frame,
                std_frame,
                x,
                y,
                std_x,
                std_y,
                photons,
                sx,
                sy,
                bg,
                lpx,
                lpy,
                ellipticity,
                net_gradient,
                n,
                n_events,
                area,
                convexhull,
                res.index.values, # group id
            ),
            dtype=CLUSTER_CENTERS_DTYPE_2D,
        )

    if hasattr(locs, "group_input"):
        group_input = _np.ndarray([_[-1] for _ in centers_])
        centers = _lib.append_to_rec(centers, group_input, "group_input")

    return centers


def cluster_center(
    grouplocs: _pd.SeriesGroupBy, 
    pixelsize: float | None = None, 
    separate_lp: bool = False,
) -> list:
    """Finds cluster centers and their attributes.

    Assumes locs to be a pandas.SeriesGroupBy object, grouped by
    cluster ids.

    Parameters
    ----------
    grouplocs : pandas.SeriesGroupBy
        Localizations grouped by cluster ids.
    pixelsize : int, optional
        Camera pixel size (used for finding volume and 3D convex hull).
        Only required for 3D localizations.
    separate_lp : bool, optional
        If True, localization precision in x and y will be calculated
        separately. Otherwise, the mean of the two is taken.

    Returns
    -------
    results : list
        Attributes used for saving the given cluster as .hdf5
        (frame, x, y, etc)
    """

    # mean and std frame
    frame = grouplocs.frame.mean()
    std_frame = grouplocs.frame.std()
    # average x and y, weighted by lpx, lpy
    x = _np.average(grouplocs.x, weights=1/(grouplocs.lpx)**2)
    y = _np.average(grouplocs.y, weights=1/(grouplocs.lpy)**2)
    std_x = grouplocs.x.std()
    std_y = grouplocs.y.std()
    # mean values
    photons = grouplocs.photons.mean()
    sx = grouplocs.sx.mean()
    sy = grouplocs.sy.mean()
    bg = grouplocs.bg.mean()
    # weighted mean loc precision
    lpx = _np.sqrt(
        error_sums_wtd(grouplocs.x, grouplocs.lpx)
        / (len(grouplocs) - 1)
    )
    lpy = _np.sqrt(
        error_sums_wtd(grouplocs.y, grouplocs.lpy)
        / (len(grouplocs) - 1)
    )
    if not separate_lp:
        lpx = (lpx + lpy) / 2
        lpy = lpx
    # other attributes
    ellipticity = sx / sy
    net_gradient = grouplocs.net_gradient.mean()
    # n_locs in cluster
    n = len(grouplocs)
    # number of binding events
    split_idx = _np.where(_np.diff(grouplocs.frame) > 3)[0] + 1 # split locs by consecutive frames
    x_events = _np.split(grouplocs.x, split_idx)
    n_events = len(x_events) # number of binding events
    if hasattr(grouplocs, "z"):
        if pixelsize is None:
            raise ValueError(
                "Camera pixel size must be specified as an integer for 3D"
                " cluster centers calculation."
            )
        z = _np.average(
            grouplocs.z, 
            weights=1/((grouplocs.lpx+grouplocs.lpy)**2),
        ) # take lpz = 2 * mean(lpx, lpy)
        std_z = grouplocs.z.std()
        # lpz = std_z
        volume = _np.power(
            (std_x + std_y + std_z / pixelsize) / 3 * 2, 3
        ) * 4.18879
        try:
            X = _np.stack(
                (grouplocs.x, grouplocs.y, grouplocs.z / pixelsize),
                axis=0,
            ).T
            hull = _ConvexHull(X)
            convexhull = hull.volume
        except:
            convexhull = 0
        result = [
            frame,
            std_frame,
            x,
            y,
            std_x,
            std_y,
            photons,
            sx,
            sy,
            bg,
            lpx,
            lpy,
            ellipticity,
            net_gradient,
            n, 
            n_events,
            z,
            std_z,
            # lpz,
            volume,
            convexhull,
        ]
    else:
        area = _np.power(std_x + std_y, 2) * _np.pi
        try:
            X = _np.stack((grouplocs.x, grouplocs.y), axis=0).T
            hull = _ConvexHull(X)
            convexhull = hull.volume
        except:
            convexhull = 0
        result = [
            frame,
            std_frame,
            x,
            y,
            std_x,
            std_y,
            photons,
            sx,
            sy,
            bg,
            lpx,
            lpy,
            ellipticity,
            net_gradient,
            n,
            n_events,
            area,
            convexhull,
        ]

    if hasattr(grouplocs, "group_input"):
        # assumes only one group input!
        result.append(_np.unique(grouplocs.group_input)[0]) 
    return result


"""
    picasso.nanotron
    ~~~~~~~~~~

    Deep learning library for classification of picked localizations

    :author: Alexander Auer, Maximilian Strauss 2020
    :copyright: Copyright (c) 2020 Jungmann Lab, MPI of Biochemistry
"""

import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm as tqdm
from scipy import ndimage
from sklearn.neural_network import MLPClassifier

from . import render, lib


def prepare_img(
    img: np.ndarray, 
    img_shape: int, 
    alpha: float = 1, 
    bg: float = 0,
) -> np.ndarray:
    """Prepare image for classification.
    
    Parameters
    ----------
    img : np.ndarray
        Input image to be prepared.
    img_shape : int
        Shape of the image (assumed to be square).
    alpha : float, optional
        Scaling factor for the image, by default 1.
    bg : float, optional
        Background value to be subtracted, by default 0.
    
    Returns
    -------
    img : np.ndarray
        Prepared image.
    """

    img = alpha * img - bg
    img = img.astype('float')
    img = img / img.max()
    img = img.clip(min=0)
    img = img.reshape(img_shape**2)

    return img


def rotate_img(img: np.ndarray, angle: float) -> np.ndarray:
    """Rotate image by a given angle.

    Parameters
    ----------
    img : np.ndarray
        Input image to be rotated.
    angle : float
        Angle in degrees by which to rotate the image.
    
    Returns
    -------
    np.ndarray
        Rotated image.
    """

    rot_img = ndimage.rotate(img, angle, reshape=False)

    return rot_img

def roi_to_img(
    locs: np.recarray, 
    pick: int, 
    radius: float, 
    oversampling: float, 
    picks: tuple[float, float] | None = None,
) -> np.ndarray:
    """Convert a region of interest (ROI) defined by localizations to an 
    image.
    
    Parameters
    ----------
    locs : np.recarray
        Localizations from which to create the image.
    pick : int
        The group number of the localizations to be picked.
    radius : float
        Radius around the mean position of the localizations to define 
        the ROI.
    oversampling : float
        Number of super-resolution pixels per camera pixel.
    picks : tuple[float, float] | None, optional
        Specific coordinates (x, y) to isolate localizations from, by
        default None. If None, uses the mean position of the localizations
        in the specified group.
        
    Returns
    -------
    pick_img : np.ndarray
        Image of the picked localizations.
    """

    # Isolate locs from pick
    pick_locs = []
    if picks is None:
        pick_locs = locs[(locs["group"] == pick)]
    else:
        x, y = picks
        pick_locs = lib.locs_at(x, y, locs, radius)
        pick_locs.sort(kind="mergesort", order="frame")
    # dirty method to avoid floating point errors with render
    radius -= 0.001

    x_mean = np.mean(pick_locs.x)
    y_mean = np.mean(pick_locs.y)

    x_min = x_mean - radius
    x_max = x_mean + radius
    y_min = y_mean - radius
    y_max = y_mean + radius

    viewport = (y_min, x_min), (y_max, x_max)

    # for debugging
    if False:
        print("mean x: {}".format(np.mean(pick_locs.x)))
        print('length x: {}'.format(x_max - x_min))
        print("mean y: {}".format(np.mean(pick_locs.y)))
        print('length y: {}'.format(y_max - y_min))
        print('radius: {}'.format(radius))
        print('viewport: {}'.format(viewport))

    # Render locs with Picasso render function
    try: 
        len_x, pick_img = render.render(pick_locs, viewport=viewport,
                                        oversampling=oversampling,
                                        blur_method='smooth')
    except:
        pass
    return pick_img


def prepare_data(
    locs: np.recarray, 
    label: int, 
    pick_radius: float,
    oversampling: float, 
    alpha: float = 10,
    bg: float = 1, 
    export: bool = False,
) -> tuple[list[np.ndarray], list[int]]:
    """Prepare data for classification by extracting images of 
    localizations. 
    
    Parameters
    ----------
    locs : np.recarray
        Localizations from which to create the images.
    label : int
        Label for the data, typically the group number.
    pick_radius : float
        Radius around the mean position of the localizations to define 
        the ROI.
    oversampling : float
        Number of super-resolution pixels per camera pixel.
    alpha : float, optional
        Scaling factor for the image, by default 10.
    bg : float, optional
        Background value to be subtracted, by default 1.
    export : bool, optional
        If True, saves the images to the './img/' directory, by default 
        False.
        
    Returns
    -------
    data : list[np.ndarray]
        List of prepared images of the localizations.
    labels : list[int]
        List of labels corresponding to the images.
    """

    img_shape = int(2 * pick_radius * oversampling)
    data = []
    labels = []

    for pick in tqdm(range(locs.group.max()), desc='Prepare '+str(label)):

        pick_img = roi_to_img(locs, pick,
                              radius=pick_radius,
                              oversampling=oversampling)

        if export is True and pick < 10:
            filename = 'label' + str(label) + '-' + str(pick)
            plt.imsave('./img/' + filename + '.png', (alpha*pick_img-bg), cmap='Greys', vmax=10)

        pick_img = prepare_img(pick_img, img_shape=img_shape, alpha=alpha, bg=bg)

        data.append(pick_img)
        labels.append(label)

    return data, label


def predict_structure(
    mlp: MLPClassifier, 
    locs: np.recarray, 
    pick: int, 
    pick_radius: float, 
    oversampling: float, 
    picks: tuple[float, float] | None = None,
) -> tuple[int, np.ndarray]:
    """Predict the structure of localizations using a trained MLP
    classifier.

    Parameters
    ----------
    mlp : MLPClassifier
        Trained MLP classifier.
    locs : np.recarray
        Localizations to predict.
    pick : int
        Index of the localizations to predict.
    pick_radius : float
        Radius around the localizations to consider.
    oversampling : float
        Oversampling factor for the image.
    picks : tuple[float, float] | None, optional
        Coordinates of the region of interest (ROI) to predict.

    Returns
    -------
    pred : int
        Predicted label for the image.
    pred_proba : np.ndarray
        Predicted probabilities for each class.
    """

    img_shape = int(2 * pick_radius * oversampling)
    img = roi_to_img(locs, pick=pick, radius=pick_radius, oversampling=oversampling, picks=picks)
    img = prepare_img(img, img_shape=img_shape, alpha=10, bg=1)
    img = img.reshape(1, img_shape**2)

    pred = mlp.predict(img)
    pred_proba = mlp.predict_proba(img)

    return pred, pred_proba


#!/usr/bin/env python
"""
    ..__main__.py
    ~~~~~~~~~~~~~~~~

    Picasso command line interface

    :authors: Joerg Schnitzbauer, Maximilian Thomas Strauss
    :copyright: Copyright (c) 2016-2019 Jungmann Lab, MPI of Biochemistry
"""

import os.path
import argparse
from . import __version__


def picasso_logo():
    """Prints the Picasso logo to the console."""

    print("    ____  _____________   __________ ____ ")
    print("   / __ \\/  _/ ____/   | / ___/ ___// __ \\")
    print("  / /_/ // // /   / /| | \\__ \\\\__ \\/ / / /")
    print(" / _____/ // /___/ ___ |___/ ___/ / /_/ / ")
    print("/_/   /___/\\____/_/  |_/____/____/\\____/  ")
    print("                                          ")


def _average(args: argparse.Namespace) -> None:
    """Run Picasso: Average.
    
    Parameters
    ----------
    iterations : int
        Number of iterations for the averaging algorithm.
    oversampling : int
        Number of super-resolution pixels per camera pixel.
    file : list of str
        List of file paths to the localization files to be averaged.
    """

    from glob import glob
    from .io import load_locs, NoMetadataFileError
    from picasso.gui import average

    kwargs = {"iterations": args.iterations, "oversampling": args.oversampling}
    paths = glob(args.file)
    if paths:
        for path in paths:
            print("Averaging {}".format(path))
            try:
                locs, info = load_locs(path)
            except NoMetadataFileError:
                continue
            kwargs["path_basename"] = os.path.splitext(path)[0] + "_avg"
            average(locs, info, **kwargs)


def _hdf2visp(path: str, pixel_size: float) -> None:
    """Convert HDF5 localization files to VISP format.

    Parameters
    ----------
    path : str
        Path to the HDF5 localization files. The file will be saved
        under the same name with a `.3d` extension.
    pixel_size : float
        Camera pixel size in nanometers.
    """

    from glob import glob

    paths = glob(path)
    if paths:
        from .io import load_locs
        import os.path
        from numpy import savetxt

        for path in paths:
            print("Converting {}".format(path))
            locs, info = load_locs(path)
            locs = locs[["x", "y", "z", "photons", "frame"]].copy()
            locs.x *= pixel_size
            locs.y *= pixel_size
            outname = os.path.splitext(path)[0] + ".3d"
            savetxt(
                outname,
                locs,
                fmt=["%.1f", "%.1f", "%.1f", "%.1f", "%d"],
                newline="\r\n",
            )


def _csv2hdf(path: str, pixelsize: float) -> None:
    """Convert CSV localization files to HDF5 format.

    Parameters
    ----------
    path : str
        Path to the CSV localization files.
    pixelsize : float
        Camera pixel size in nanometers.
    """

    from glob import glob
    from tqdm import tqdm as _tqdm
    import pandas as pd

    paths = glob(path)
    if paths:
        from .io import save_locs
        import os.path
        import numpy as _np

        for path in _tqdm(paths):
            print("Converting {}".format(path))
            data = pd.read_csv(path)

            try:
                frames = data["frame"].astype(int)
                # make sure frames start at zero:
                frames = frames - _np.min(frames)
                x = data["x [nm]"] / pixelsize
                y = data["y [nm]"] / pixelsize
                photons = data["intensity [photon]"].astype(int)

                bg = data["offset [photon]"].astype(int)
                lpx = data["uncertainty_xy [nm]"] / pixelsize
                lpy = data["uncertainty_xy [nm]"] / pixelsize

                if "z_nm" in list(data):
                    # TODO update other column labels
                    z = data["z_nm"] / pixelsize
                    sx = data["sigma1_nm"] / pixelsize
                    sy = data["sigma2_nm"] / pixelsize

                    LOCS_DTYPE = [
                        ("frame", "u4"),
                        ("x", "f4"),
                        ("y", "f4"),
                        ("z", "f4"),
                        ("photons", "f4"),
                        ("sx", "f4"),
                        ("sy", "f4"),
                        ("bg", "f4"),
                        ("lpx", "f4"),
                        ("lpy", "f4"),
                    ]

                    locs = _np.rec.array(
                        (frames, x, y, z, photons, sx, sy, bg, lpx, lpy),
                        dtype=LOCS_DTYPE,
                    )

                else:
                    sx = data["sigma [nm]"] / pixelsize
                    sy = data["sigma [nm]"] / pixelsize

                    LOCS_DTYPE = [
                        ("frame", "u4"),
                        ("x", "f4"),
                        ("y", "f4"),
                        ("photons", "f4"),
                        ("sx", "f4"),
                        ("sy", "f4"),
                        ("bg", "f4"),
                        ("lpx", "f4"),
                        ("lpy", "f4"),
                    ]

                    locs = _np.rec.array(
                        (frames, x, y, photons, sx, sy, bg, lpx, lpy),
                        dtype=LOCS_DTYPE,
                    )

                locs.sort(kind="mergesort", order="frame")

                img_info = {}
                img_info["Generated by"] = f"Picasso v{__version__} csv2hdf"
                img_info["Frames"] = int(_np.max(frames)) + 1
                img_info["Height"] = int(_np.ceil(_np.max(y)))
                img_info["Width"] = int(_np.ceil(_np.max(x)))

                info = []
                info.append(img_info)

                base, ext = os.path.splitext(path)
                out_path = base + "_locs.hdf5"
                save_locs(out_path, locs, info)
                print("Saved to {}.".format(out_path))
            except Exception as e:
                print(e)
                print("Error. Datatype not understood.")
                raise e


def _hdf2csv(path: str) -> None:
    """Convert HDF5 localization files to CSV format."""

    from glob import glob
    import pandas as pd
    from tqdm import tqdm as _tqdm
    from os.path import isdir

    if isdir(path):
        paths = glob(path + "/*.hdf5")
    else:
        paths = glob(path)
    if paths:
        import os.path
        from .io import load_locs

        for path in _tqdm(paths):
            base, ext = os.path.splitext(path)
            if ext == ".hdf5":
                print("Converting {}".format(path))
                out_path = base + ".csv"
                locs = load_locs(path)[0]
                df = pd.DataFrame(locs)
                print("A total of {} rows loaded".format(len(locs)))
                df.to_csv(out_path, sep=",", encoding="utf-8")
    print("Complete.")


def _link(files: str, d_max: float, tolerance: float) -> None:
    """Link localizations in HDF5 files, see ``postprocess.link`` for
    details."""

    import numpy as _np
    from tqdm import tqdm as _tqdm
    from . import lib as _lib
    from h5py import File

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, postprocess

        for path in paths:
            try:
                locs, info = io.load_locs(path)
            except io.NoMetadataFileError:
                continue
            linked_locs = postprocess.link(locs, info, d_max, tolerance)
            base, ext = os.path.splitext(path)
            link_info = {
                "Maximum Distance": d_max,
                "Maximum Transient Dark Time": tolerance,
                "Generated by": f"Picasso v{__version__} Link",
            }
            info.append(link_info)
            io.save_locs(base + "_link.hdf5", linked_locs, info)

            try:
                # Check if there is a _clusters.hdf5 file present
                # if yes update this file
                cluster_path = base[:-7] + "_clusters.hdf5"
                print(cluster_path)
                clusters = io.load_clusters(cluster_path)
                print("Clusterfile detected. Updating entries.")

                n_after_link = []
                linked_len = []
                linked_n = []
                linked_photonrate = []

                for group in _tqdm(_np.unique(clusters["groups"])):
                    temp = linked_locs[linked_locs["group"] == group]
                    if len(temp) > 0:
                        n_after_link.append(len(temp))
                        linked_len.append(_np.mean(temp["len"]))
                        linked_n.append(_np.mean(temp["n"]))
                        linked_photonrate.append(_np.mean(temp["photon_rate"]))

                clusters = _lib.append_to_rec(
                    clusters,
                    _np.array(n_after_link, dtype=_np.int32),
                    "n_after_link",
                )
                clusters = _lib.append_to_rec(
                    clusters,
                    _np.array(linked_len, dtype=_np.int32),
                    "linked_len",
                )
                clusters = _lib.append_to_rec(
                    clusters, _np.array(linked_n, dtype=_np.int32), "linked_n"
                )
                clusters = _lib.append_to_rec(
                    clusters,
                    _np.array(linked_photonrate, dtype=_np.float32),
                    "linked_photonrate",
                )
                with File(cluster_path, "w") as clusters_file:
                    clusters_file.create_dataset("clusters", data=clusters)
            except Exception as e:
                print("No clusterfile found for updating.")
                continue


def _cluster_combine(files: str) -> None:
    """Combine clusters in HDF5 files. See 
    ``postprocess.cluster_combine`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, postprocess

        for path in paths:
            try:
                locs, info = io.load_locs(path)
            except io.NoMetadataFileError:
                continue
            combined_locs = postprocess.cluster_combine(locs)
            base, ext = os.path.splitext(path)
            combined_info = {"Generated by": f"Picasso v{__version__} Combine"}
            info.append(combined_info)
            io.save_locs(base + "_comb.hdf5", combined_locs, info)


def _cluster_combine_dist(files: str) -> None:
    """Combine clusters in HDF5 files based on distance. See
    ``postprocess.cluster_combine_dist`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, postprocess

        for path in paths:
            try:
                locs, info = io.load_locs(path)
            except io.NoMetadataFileError:
                continue
            combinedist_locs = postprocess.cluster_combine_dist(locs)
            base, ext = os.path.splitext(path)
            cluster_combine_dist_info = {"Generated by": f"Picasso v{__version__} Combineidst"}
            info.append(cluster_combine_dist_info)
            io.save_locs(base + "_cdist.hdf5", combinedist_locs, info)


def _clusterfilter(
    files: str,
    clusterfile: str,
    parameter: str,
    minval: float,
    maxval: float,
) -> None:
    """Filter localizations based on cluster parameters.

    Parameters
    ----------
    files : str
        Glob pattern for input files.
    clusterfile : str
        Path to the cluster file.
    parameter : str
        Name of the parameter to filter on.
    minval : float
        Minimum value for the parameter.
    maxval : float
        Maximum value for the parameter.
    """

    from glob import glob
    from tqdm import tqdm
    import numpy as np

    paths = glob(files)
    if paths:
        from . import io

        for path in paths:
            try:
                locs, info = io.load_locs(path)
            except io.NoMetadataFileError:
                continue

            clusters = io.load_clusters(clusterfile)
            try:
                selector = (clusters[parameter] > minval) & (
                    clusters[parameter] < maxval
                )
                if np.sum(selector) == 0:
                    print("Error: No localizations in range. Filtering aborted.")
                elif np.sum(selector) == len(selector):
                    print("Error: All localizations in range. Filtering aborted.")
                else:
                    print("Isolating locs.. Step 1: in range")
                    groups = clusters["groups"][selector]
                    first = True
                    for group in tqdm(groups):
                        if first:
                            all_locs = locs[locs["group"] == group]
                            first = False
                        else:
                            all_locs = np.append(all_locs, locs[locs["group"] == group])

                    base, ext = os.path.splitext(path)
                    clusterfilter_info = {
                        "Generated by": f"Picasso v{__version__} Clusterfilter - in",
                        "Paramter": parameter,
                        "Minval": minval,
                        "Maxval": maxval,
                    }
                    info.append(clusterfilter_info)
                    all_locs.sort(kind="mergesort", order="frame")
                    all_locs = all_locs.view(np.recarray)
                    out_path = base + "_filter_in.hdf5"
                    io.save_locs(out_path, all_locs, info)
                    print("Complete. Saved to: {}".format(out_path))

                    print("Isolating locs.. Step 2: out of range")
                    groups = clusters["groups"][~selector]
                    first = True
                    for group in tqdm(groups):
                        if first:
                            all_locs = locs[locs["group"] == group]
                            first = False
                        else:
                            all_locs = np.append(all_locs, locs[locs["group"] == group])

                    base, ext = os.path.splitext(path)
                    clusterfilter_info = {
                        "Generated by": f"Picasso v{__version__} Clusterfilter - out",
                        "Paramter": parameter,
                        "Minval": minval,
                        "Maxval": maxval,
                    }
                    info.append(clusterfilter_info)
                    all_locs.sort(kind="mergesort", order="frame")
                    all_locs = all_locs.view(np.recarray)
                    out_path = base + "_filter_out.hdf5"
                    io.save_locs(out_path, all_locs, info)
                    print("Complete. Saved to: {}".format(out_path))

            except ValueError:
                print("Error: Field {} not found.".format(parameter))


def _undrift(
    files: str, 
    segmentation: int, 
    display: bool = True, 
    fromfile: str | None = None,
) -> None:
    """Runs RCC undrifting on the given files. See 
    ``postprocess.undrift`` for details. Alternatively, it can read the
    drift .txt file to apply the drift correction."""

    import glob
    from . import io, postprocess
    from numpy import genfromtxt, savetxt

    paths = glob.glob(files)
    undrift_info = {"Generated by": f"Picasso v{__version__} Undrift"}
    if fromfile is not None:
        undrift_info["From File"] = fromfile
        drift = genfromtxt(fromfile)
    else:
        undrift_info["Segmentation"] = segmentation
    for path in paths:
        try:
            locs, info = io.load_locs(path)
        except io.NoMetadataFileError:
            continue
        if fromfile is not None:
            # this works for mingjies drift files but not for the own ones
            locs.x -= drift[:, 1][locs.frame]
            locs.y -= drift[:, 0][locs.frame]
            if display:
                import matplotlib.pyplot as plt

                plt.style.use("ggplot")
                plt.figure(figsize=(17, 6))
                plt.suptitle("Estimated drift")
                plt.subplot(1, 2, 1)
                plt.plot(drift[:, 1], label="x")
                plt.plot(drift[:, 0], label="y")
                plt.legend(loc="best")
                plt.xlabel("Frame")
                plt.ylabel("Drift (pixel)")
                plt.subplot(1, 2, 2)
                plt.plot(
                    drift[:, 1],
                    drift[:, 0],
                    color=list(plt.rcParams["axes.prop_cycle"])[2]["color"],
                )
                plt.axis("equal")
                plt.xlabel("x")
                plt.ylabel("y")
                plt.show()
        else:
            print("Undrifting file {}".format(path))
            drift, locs = postprocess.undrift(locs, info, segmentation, display=display)

            undrift_info["Drift X"] = float(drift["x"].mean())
            undrift_info["Drift Y"] = float(drift["y"].mean())

        info.append(undrift_info)
        base, ext = os.path.splitext(path)
        io.save_locs(base + "_undrift.hdf5", locs, info)
        savetxt(base + "_drift.txt", drift, header="dx\tdy", newline="\r\n")


def _undrift_aim(
    files: str, 
    segmentation: int, 
    intersectdist: float = 20/130, 
    roiradius: float = 60/130
) -> None:
    """Runs AIM undrifting on the given files. See ``aim.aim`` for 
    details."""

    import glob
    from . import io, aim
    from numpy import savetxt

    paths = glob.glob(files)
    for path in paths:
        try:
            locs, info = io.load_locs(path)
        except io.NoMetadataFileError:
            continue
        print("Undrifting file {}".format(path))
        locs, new_info, drift = aim.aim(locs, info, segmentation, intersectdist, roiradius)
        base, ext = os.path.splitext(path)
        io.save_locs(base + "_aim.hdf5", locs, new_info)
        savetxt(base + "_aimdrift.txt", drift, header="dx\tdy", newline="\r\n")  


def _density(files: str, radius: float) -> None:
    """Compute local density of localizations in HDF5 files. See
    ``postprocess.compute_local_density`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, postprocess

        for path in paths:
            locs, info = io.load_locs(path)
            locs = postprocess.compute_local_density(locs, info, radius)
            base, ext = os.path.splitext(path)
            density_info = {
                "Generated by": f"Picasso v{__version__} Density",
                "Radius": radius,
            }
            info.append(density_info)
            io.save_locs(base + "_density.hdf5", locs, info)


def _dbscan(
    files: str, 
    radius: float, 
    min_density: float, 
    pixelsize: float | None = None,
) -> None:
    """Run DBSCAN clustering on localizations in HDF5 files. See
    ``clusterer.dbscan`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, clusterer

        for path in paths:
            print("Loading {} ...".format(path))
            locs, info = io.load_locs(path)
            locs = clusterer.dbscan(locs, radius, min_density, pixelsize)
            clusters = clusterer.find_cluster_centers(locs, pixelsize)
            base, _ = os.path.splitext(path)
            dbscan_info = {
                "Generated by": f"Picasso v{__version__} DBSCAN",
                "Radius": radius,
                "Minimum local density": min_density,
            }
            info.append(dbscan_info)
            io.save_locs(base + "_dbscan.hdf5", locs, info)
            io.save_locs(base + "_dbclusters.hdf5", clusters, info)
            print(
                "Clustering executed. Results are saved in: \n"
                f"{base}_dbscan.hdf5\n"
                f"{base}_dbclusters.hdf5"
            )


def _hdbscan(
    files: str, 
    min_cluster: int, 
    min_samples: int, 
    pixelsize: float | None = None,
) -> None:
    """Run HDBSCAN clustering on localizations in HDF5 files. See
    ``clusterer.hdbscan`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, clusterer

        for path in paths:
            print("Loading {} ...".format(path))
            locs, info = io.load_locs(path)
            locs = clusterer.hdbscan(locs, min_cluster, min_samples, pixelsize)
            clusters = clusterer.find_cluster_centers(locs, pixelsize)
            base, ext = os.path.splitext(path)
            hdbscan_info = {
                "Generated by": f"Picasso v{__version__} HDBSCAN",
                "Min. cluster": min_cluster,
                "Min. samples": min_samples,
            }
            info.append(hdbscan_info)
            io.save_locs(base + "_hdbscan.hdf5", locs, info)
            io.save_locs(base + "_hdbclusters.hdf5", clusters, info)
            print(
                "Clustering executed. Results are saved in: \n"
                f"{base}_hdbscan.hdf5\n"
                f"{base}_hdbclusters.hdf5"
            )

def _smlm_clusterer(
    files: str, 
    radius: float, 
    min_locs: int, 
    pixelsize: float | None = None, 
    basic_fa: bool = False, 
    radius_z: float | None = None,
) -> None:
    """Run SMLM clustering on localizations in HDF5 files. See
    ``clusterer.cluster`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, clusterer
        params = {
            "radius_xy": radius,
            "radius_z": radius_z,
            "min_locs": min_locs,
            "frame_analysis": basic_fa,
        }
        for path in paths:
            print("Loading {} ...".format(path))
            locs, info = io.load_locs(path)
            locs = clusterer.cluster(locs, **params, pixelsize=pixelsize)
            clusters = clusterer.find_cluster_centers(locs, pixelsize)
            base, ext = os.path.splitext(path)
            smlm_cluster_info = {
                "Generated by": f"Picasso v{__version__} SMLM clusterer",
                "Radius_xy": radius,
                "Radius_z": radius_z,
                "Min locs": min_locs,
                "Basic frame analysis": basic_fa,
            }
            info.append(smlm_cluster_info)
            io.save_locs(base + "_clusters.hdf5", locs, info)
            io.save_locs(base + "_cluster_centers.hdf5", clusters, info)
            print(
                "Clustering executed. Results are saved in: \n"
                f"{base}_clusters.hdf5\n"
                f"{base}_cluster_centers.hdf5"
            )


def _nneighbor(files: str) -> None:
    """Calculate the minimum distance to the nearest neighbor for each
    localization in the given HDF5 files. The results are saved in a
    text file with the same name as the input file, but with a
    `_minval.txt` suffix. The distances are calculated using the
    Euclidean distance metric."""

    import glob
    import h5py as _h5py
    import numpy as np
    from scipy.spatial import distance

    paths = glob.glob(files)
    if paths:
        for path in paths:
            print("Loading {} ...".format(path))
            with _h5py.File(path, "r") as locs_file:
                locs = locs_file["clusters"][...]
            clusters = np.rec.array(locs, dtype=locs.dtype)
            points = np.array(clusters[["com_x", "com_y"]].tolist())
            alldist = distance.cdist(points, points)
            alldist[alldist == 0] = float("inf")
            minvals = np.amin(alldist, axis=0)
            base, ext = os.path.splitext(path)
            out_path = base + "_minval.txt"
            np.savetxt(out_path, minvals, newline="\r\n")
            print("Saved filest o: {}".format(out_path))


def _dark(files: str) -> None:
    """Compute dark times for localizations in HDF5 files. See
    ``postprocess.compute_dark_times`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from . import io, postprocess

        for path in paths:
            locs, info = io.load_locs(path)
            locs = postprocess.compute_dark_times(locs)
            base, ext = os.path.splitext(path)
            d_info = {"Generated by": f"Picasso v{__version__} Dark"}
            info.append(d_info)
            io.save_locs(base + "_dark.hdf5", locs, info)


def _align(files: str, display: bool) -> None:
    """Aligns localization files using RCC, see ``postprocess.align`` 
    for details."""

    from glob import glob
    from itertools import chain
    from .io import load_locs, save_locs
    from .postprocess import align
    from os.path import splitext

    files = list(chain(*[glob(_) for _ in files]))
    print("Aligning files:")
    for f in files:
        print("  " + f)
    locs_infos = [load_locs(_) for _ in files]
    locs = [_[0] for _ in locs_infos]
    infos = [_[1] for _ in locs_infos]
    aligned_locs = align(locs, infos, display=display)
    align_info = {"Generated by": f"Picasso v{__version__} Align", "Files": files}
    for file, locs_, info in zip(files, aligned_locs, infos):
        info.append(align_info)
        base, ext = splitext(file)
        save_locs(base + "_align.hdf5", locs_, info)


def _join(files: list[str], keep_index: bool = True) -> None:
    """Join multiple localization files into one."""

    from .io import load_locs, save_locs
    from os.path import splitext
    from numpy import append
    import numpy as np

    locs, info = load_locs(files[0])
    total_frames = info[0]["Frames"]
    join_info = {"Generated by": f"Picasso v{__version__} Join", "Files": [files[0]]}
    for path in files[1:]:
        locs_, info_ = load_locs(path)
        try:
            n_frames = info[0]["Frames"]
            total_frames += n_frames
            if not keep_index:
                locs_["frame"] += total_frames - n_frames
            locs = append(locs, locs_)
            join_info["Files"].append(path)
        except TypeError:
            print(
                "An error occured.\n"
                "Unable to join files."
                " Make sure they have the same columns."
            )
    base, ext = splitext(files[0])
    info.append(join_info)
    if not keep_index:
        info[0]["Frames"] = total_frames
    locs.sort(kind="mergesort", order="frame")
    locs = locs.view(np.recarray)
    save_locs(base + "_join.hdf5", locs, info)


def _groupprops(files: str) -> None:
    """Calculate group properties for localizations in HDF5 files.
    See ``postprocess.groupprops`` for details."""

    import glob

    paths = glob.glob(files)
    if paths:
        from .io import load_locs, save_datasets
        from .postprocess import groupprops
        from os.path import splitext

        for path in paths:
            locs, info = load_locs(path)
            groups = groupprops(locs)
            base, ext = splitext(path)
            save_datasets(base + "_groupprops.hdf5", info, locs=locs, groups=groups)


def _pair_correlation(files: str, bin_size: float, r_max: float) -> None:
    """Calculate pair-correlation for localizations in HDF5 files. See
    ``postprocess.pair_correlation`` for details."""

    from glob import glob

    paths = glob(files)
    if paths:
        from .io import load_locs
        from .postprocess import pair_correlation
        from matplotlib.pyplot import plot, style, show, xlabel, ylabel, title

        style.use("ggplot")
        for path in paths:
            print("Loading {}...".format(path))
            locs, info = load_locs(path)
            print("Calculating pair-correlation...")
            bins_lower, pc = pair_correlation(locs, info, bin_size, r_max)
            plot(bins_lower - bin_size / 2, pc)
            xlabel("r (pixel)")
            ylabel("pair-correlation (pixel^-2)")
            title("Pair-correlation. Bin size: {}, R max: {}".format(bin_size, r_max))
            show()


def _start_server() -> None:
    """Start the Streamlit server for the Picasso GUI."""

    import os
    import sys
    from streamlit.web import cli as stcli

    print("                                          ")
    picasso_logo()
    print("                 server")
    print("                                          ")

    HOME = os.path.expanduser("~")

    ST_PATH = os.path.join(HOME, ".streamlit")

    for folder in [ST_PATH]:
        if not os.path.isdir(folder):
            os.mkdir(folder)

    _this_file = os.path.abspath(__file__)
    _this_dir = os.path.dirname(_this_file)

    file_path = os.path.join(_this_dir, "server", "app.py")

    # Check if streamlit credentials exists
    ST_CREDENTIALS = os.path.join(ST_PATH, "credentials.toml")
    if not os.path.isfile(ST_CREDENTIALS):
        with open(ST_CREDENTIALS, "w") as file:
            file.write("[general]\n")
            file.write('\nemail = ""')

    theme = []

    theme.append("--theme.backgroundColor=#FFFFFF")
    theme.append("--theme.secondaryBackgroundColor=#f0f2f6")
    theme.append("--theme.textColor=#262730")
    theme.append("--theme.font=sans serif")
    theme.append("--theme.primaryColor=#18212b")

    args = [
        "streamlit",
        "run",
        file_path,
        "--global.developmentMode=false",
        "--server.port=8501",
        "--browser.gatherUsageStats=False",
    ]

    # args.extend(theme)

    sys.argv = args

    sys.exit(stcli.main())


def _nanotron(args: argparse.Namespace) -> None:
    """Run nanoTRON prediction on localization files.
    
    Parameters
    ----------
    files : str
        Path to the localization files or a directory containing
        HDF5 files.
    model_path : str
        Path to the nanoTRON model file.
    model_pth : str
        Path to the nanoTRON model weights file.
    """

    from glob import glob
    from os.path import isdir
    from .io import load_locs, NoMetadataFileError
    from picasso.gui import nanotron

    files = args.files
    model = args.model_path
    kwargs = {"model": args.model_pth}

    if isdir(files):
        print("Analyzing folder")
        paths = glob(files + "/*.hdf5")
        print("A total of {} files detected".format(len(paths)))
    else:
        paths = glob(files)

    if paths:
        for path in paths:
            print("nanoTRON predicting {}".format(path))
            try:
                locs, info = load_locs(path)
            except NoMetadataFileError:
                continue
            raise NotImplementedError  # Todo: Include call to proper prediction routine
            # predict(locs, info, **kwargs)


def _localize(args: argparse.Namespace) -> None:
    """Localize molecules in microscopy images.
    
    Parameters
    ----------
    files : str
        Path to the microscopy image files or a directory containing
        image files.
    fit_method : str
        Method to use for fitting localizations. Options are:
        - 'mle': Maximum Likelihood Estimation
        - 'lq-3d': LQ 3D fitting
        - 'lq-gpu-3d': LQ GPU 3D fitting
    box_side_length : int
        Side length of the box used for localization.
    gradient : float
        Minimum net gradient for localization.
    roi : list of int
        Region of interest defined as [y_min, x_min, y_max, x_max].
    baseline : float
        Baseline value for the camera.
    sensitivity : float
        Sensitivity of the camera.
    gain : float
        Gain of the camera.
    qe : float
        Not used in the calculations.
    """

    files = args.files
    from glob import glob
    from .io import load_movie, save_locs, save_info
    from .localize import (
        get_spots,
        identify_async,
        identifications_from_futures,
        fit_async,
        locs_from_fits,
        add_file_to_db,
    )
    from os.path import splitext, isdir
    from time import sleep
    from . import gausslq, avgroi
    import os.path as _ospath
    import re as _re
    import os as _os
    import yaml as yaml

    picasso_logo()
    print("Localize - Parameters:")
    print("{:<8} {:<15} {:<10}".format("No", "Label", "Value"))

    if args.fit_method == "lq-gpu":
        if gausslq.gpufit_installed:
            print("GPUfit installed")
        else:
            raise Exception("GPUfit not installed. Aborting.")

    for index, element in enumerate(vars(args)):
        try:
            print("{:<8} {:<15} {:<10}".format(index + 1, element, getattr(args, element)))
        except TypeError: # if None is default value
            print("{:<8} {:<15} {}".format(index + 1, element, "None"))
    print("------------------------------------------")

    def check_consecutive_tif(filepath):
        """
        Function to only return the first file of a consecutive ome.tif series
        to not reconstruct all of them as load_movie automatically detects
        consecutive files. E.g. have a folder with file.ome.tif,
        file_1.ome.tif, file_2.ome.tif, will return only file.ome.tif
        Or NDTiffStacks where files have format file.tif, file_1.tif, etc.
        """
        files = glob(filepath + "/*.tif")
        newlist = [_ospath.abspath(file) for file in files]
        for file in files:
            path = _ospath.abspath(file)
            directory = _ospath.dirname(path)
            if "NDTiffStack" in path:
                base, ext = _ospath.splitext(path)
                base = _re.escape(base)
                pattern = _re.compile(base + r"_(\d*).tif")
            else:
                base, ext = _ospath.splitext(
                    _ospath.splitext(path)[0]
                )  # split two extensions as in .ome.tif
                base = _re.escape(base)
                # This matches the basename + an appendix of the file number
                pattern = _re.compile(base + r"_(\d*).ome.tif")  
            entries = [_.path for _ in _os.scandir(directory) if _.is_file()]
            matches = [_re.match(pattern, _) for _ in entries]
            matches = [_ for _ in matches if _ is not None]
            datafiles = [_.group(0) for _ in matches]
            if datafiles != []:
                for element in datafiles:
                    newlist.remove(element)
        return newlist

    if isdir(files):
        print("Analyzing folder")

        tif_files = check_consecutive_tif(files)

        paths = tif_files + glob(files + "/*.raw") + glob(files + "/*.nd2")
        print("A total of {} files detected".format(len(paths)))
    else:
        paths = glob(files)

    # Check for raw files: make sure that each contains a yaml file
    def prompt_info():
        info = {}
        info["Byte Order"] = input("Byte Order (< or >): ")
        info["Data Type"] = input('Data Type (e.g. "uint16"): ')
        info["Frames"] = int(input("Frames: "))
        info["Height"] = int(input("Height: "))
        info["Width"] = int(input("Width: "))
        save = input("Use for all remaining raw files in folder (y/n)?") == "y"
        return info, save

    save = False
    for path in paths:
        base, ext = _ospath.splitext(path)
        if ext == ".raw":
            if not _os.path.isfile(base + ".yaml"):
                print("No yaml found for {}. Please enter:".format(path))
                if not save:
                    info, save = prompt_info()
                info_path = base + ".yaml"
                save_info(info_path, [info])

    if paths:
        print(args)
        box = args.box_side_length
        min_net_gradient = args.gradient
        roi = args.roi
        if roi is not None:
            y_min, x_min, y_max, x_max = roi
            roi = [[y_min, x_min], [y_max, x_max]]
        camera_info = {}
        camera_info["Baseline"] = args.baseline
        camera_info["Sensitivity"] = args.sensitivity
        camera_info["Gain"] = args.gain
        camera_info["Qe"] = args.qe

        if args.fit_method == "mle":
            # use default settings
            convergence = 0.001
            max_iterations = 1000
        else:
            convergence = 0
            max_iterations = 0

        if args.fit_method == "lq-3d" or args.fit_method == "lq-gpu-3d":
            from . import zfit

            print("------------------------------------------")
            print("Fitting 3D")

            if not os.path.isfile(args.zc):
                print(
                    "Given path for calibration file not found. Please enter manually:"
                )
                zpath = input("Path to *.yaml calibration file: ")
            else:
                zpath = args.zc

            if args.mf == 0:
                magnification_factor = float(input("Enter Magnification factor: "))
            else:
                magnification_factor = args.mf

            try:
                with open(zpath, "r") as f:
                    z_calibration = yaml.full_load(f)
            except Exception as e:
                print(e)
                print("Error loading calibration file.")
                raise

        for i, path in enumerate(paths):
            print("------------------------------------------")
            print("------------------------------------------")
            print("Processing {}, File {} of {}".format(path, i + 1, len(paths)))
            print("------------------------------------------")
            movie, info = load_movie(path)
            current, futures = identify_async(movie, min_net_gradient, box, roi=roi)
            n_frames = len(movie)
            while current[0] < n_frames:
                print(
                    "Identifying in frame {:,} of {:,}".format(
                        current[0] + 1, n_frames
                    ),
                    end="\r",
                )
                sleep(0.2)
            print("Identifying in frame {:,} of {:,}".format(n_frames, n_frames))
            ids = identifications_from_futures(futures)

            if args.fit_method == "lq" or args.fit_method == "lq-3d":
                spots = get_spots(movie, ids, box, camera_info)
                theta = gausslq.fit_spots_parallel(spots, asynch=False)
                locs = gausslq.locs_from_fits(ids, theta, box, args.gain)
            elif args.fit_method == "lq-gpu" or args.fit_method == "lq-gpu-3d":
                spots = get_spots(movie, ids, box, camera_info)
                theta = gausslq.fit_spots_gpufit(spots)
                em = camera_info["Gain"] > 1
                locs = gausslq.locs_from_fits_gpufit(ids, theta, box, em)
            elif args.fit_method == "mle":
                current, thetas, CRLBs, likelihoods, iterations = fit_async(
                    movie, camera_info, ids, box, convergence, max_iterations
                )
                n_spots = len(ids)
                while current[0] < n_spots:
                    print(
                        "Fitting spot {:,} of {:,}".format(current[0] + 1, n_spots),
                        end="\r",
                    )
                    sleep(0.2)
                print("Fitting spot {:,} of {:,}".format(n_spots, n_spots))
                locs = locs_from_fits(ids, thetas, CRLBs, likelihoods, iterations, box)

            elif args.fit_method == "avg":
                spots = get_spots(movie, ids, box, camera_info)
                theta = avgroi.fit_spots_parallel(spots, asynch=False)
                locs = avgroi.locs_from_fits(ids, theta, box, args.gain)

            else:
                print("This should never happen...")

            try:
                px = args.pixelsize
            except Exception as e:
                px = None

            localize_info = {
                "Generated by": f"Picasso v{__version__} Localize",
                "ROI": None, #TODO: change if ROI is given
                "Box Size": box,
                "Min. Net Gradient": min_net_gradient,
                "Pixelsize": px,
                "Fit method": args.fit_method
            }
            localize_info.update(camera_info)
            if args.fit_method == "mle":
                localize_info["Convergence Criterion"] = convergence
                localize_info["Max. Iterations"] = max_iterations

            if args.fit_method == "lq-3d" or args.fit_method == "lq-gpu-3d":
                print("------------------------------------------")
                print("Fitting 3D...", end="")
                fs = zfit.fit_z_parallel(
                    locs,
                    info,
                    z_calibration,
                    magnification_factor,
                    filter=0,
                    asynch=True,
                )
                locs = zfit.locs_from_futures(fs, filter=0)
                localize_info["Z Calibration Path"] = zpath
                localize_info["Z Calibration"] = z_calibration
                print("complete.")
                print("------------------------------------------")

            info.append(localize_info)
            info.append(camera_info)

            base, ext = splitext(path)

            try:
                sfx = args.suffix
            except Exception as e:
                sfx = ""

            out_path = f"{base}{sfx}_locs.hdf5"
            save_locs(out_path, locs, info)
            print("File saved to {}".format(out_path))

            if hasattr(args, "database"):
                CHECK_DB = args.database
            else:
                CHECK_DB = False

            if CHECK_DB:
                print("\n")
                print("Assesing quality and adding to DB")
                add_file_to_db(path, out_path)
                print("Done.")
                print("\n")

            if args.drift > 0:
                print("Undrifting file:")
                print("------------------------------------------")
                try:
                    _undrift(out_path, args.drift, display=False, fromfile=None)
                except Exception as e:
                    print(e)
                    print("Drift correction failed for {}".format(out_path))

            print("                                          ")
    else:
        print("Error. No files found.")
        raise FileNotFoundError


def _render(args: argparse.Namespace) -> None:
    """Render localization files to images.
    
    Parameters
    ----------
    files : str
        Path to the localization files or a directory containing
        HDF5 files.
    oversampling : int
        Number of super-resolution pixels per camera pixel.
    blur_method : str 
        Defines localizations' blur. The string has to be one of 
        'gaussian', 'gaussian_iso', 'smooth', 'convolve'. If None, no
        blurring is applied.
    min_blur_width : float
        Minimum width of the blur kernel in pixels.
    vmin : float
        Minimum value for the color scale.
    vmax : float
        Maximum value for the color scale.  
    cmap : str
        Colormap to use for rendering. If None, the colormap from
        user settings is used.
    scaling : str
        If 'yes', the image is scaled to the range [vmin, vmax].
        If 'no', the image is not scaled.
    silent : bool
        If True, the rendered images are not opened automatically.
    """

    from .lib import locs_glob_map
    from .render import render
    from os.path import splitext
    from matplotlib.pyplot import imsave
    from os import startfile
    from os.path import isdir
    from .io import load_user_settings, save_user_settings
    from tqdm import tqdm
    from glob import glob

    def render_many(
        locs,
        info,
        path,
        oversampling,
        blur_method,
        min_blur_width,
        vmin,
        vmax,
        scaling,
        cmap,
        silent,
    ):
        if blur_method == "none":
            blur_method = None
        N, image = render(
            locs,
            info,
            oversampling,
            blur_method=blur_method,
            min_blur_width=min_blur_width,
        )
        base, ext = splitext(path)
        out_path = base + ".png"
        im_max = image.max() / 100
        if scaling == "yes":
            imsave(out_path, image, vmin=vmin * im_max, vmax=vmax * im_max, cmap=cmap)
        else:
            imsave(out_path, image, vmin=vmin, vmax=vmax, cmap=cmap)
        if not silent:
            startfile(out_path)

    settings = load_user_settings()
    cmap = args.cmap
    if cmap is None:
        try:
            cmap = settings["Render"]["Colormap"]
        except KeyError:
            cmap = "viridis"
    settings["Render"]["Colormap"] = cmap
    save_user_settings(settings)

    if isdir(args.files):
        print("Analyzing folder")
        paths = glob(args.files + "/*.hdf5")
        print("A total of {} files detected. Rendering.".format(len(paths)))

        for path in tqdm(paths):
            locs_glob_map(
                render_many,
                path,
                args=(
                    args.oversampling,
                    args.blur_method,
                    args.min_blur_width,
                    args.vmin,
                    args.vmax,
                    args.scaling,
                    cmap,
                    True,
                ),
            )

    else:
        locs_glob_map(
            render_many,
            args.files,
            args=(
                args.oversampling,
                args.blur_method,
                args.min_blur_width,
                args.vmin,
                args.vmax,
                args.scaling,
                cmap,
                args.silent,
            ),
        )


def _spinna_batch_analysis(
    parameters_filename: str, 
    asynch: bool = True, 
    bootstrap: bool = False, 
    verbose: bool = False,
) -> None:
    """SPINNA batch analysis. Results are automatically saved in the
    a new subfolder named "parameters_filename_fitting_results" in the 
    folder where the parameters file is located. Parameters should be 
    provided in .csv file. Each row in the file specifies one analysis 
    run. The parameters (columns) are:

    - "structures_filename" : Name of the files with structures saved 
        (.yaml).
    - "exp_data_TARGET" : Name of the file with experimental data 
        (.hdf5). Each target in the structures must have a 
        corresponding column, for example, "exp_data_EGFR".
    - "le_TARGET" : Labeling efficiency (%) for each target. Each 
        target in the structures must have a corresponding column, 
        for example, "le_EGFR".
    - "label_unc_TARGET" : Label uncertainty (nm) for each target. Each
        target in the structures must have a corresponding column, 
        for example: "label_unc_EGFR".
    - "granularity" : Granularity used in parameters search space
        generation. The higher the value the more combinations of 
        structure counts will be tested.
    - "sim_repeats" : Number of simulation repeats used for obtaining
        smoother NND histograms.
    - "save_filename" : Name of the file where the results will be
        saved. 
    - "NND_bin" : Bin size (nm) for the nearest neighbor distance (NND)
        histogram (plotting only).
    - "NND_maxdist" : Maximum distance (nm) for the nearest neighbor
        distance (NND) histogram (plotting only).
    
    Depending on whether a homo- or heterogeneous (masked) distribution 
    is used, the following columns must be present:
    * For homogeneous distribution:

    - "area" or "volume" : Area (2D simulation) or volume (3D 
        simulation) of the simulated ROI (um^2 or um^3).
    - "z_range" : Applicable only when "volume" is provided. Defines
        the range of z coordinates (nm) of simulated molecular targets.
    
    * For heterogeneous distribution:
    - "mask_filename_TARGET" : Name of the .npy file with the mask saved for 
        each molecular target. Each target in the structures must have 
        a corresponding column, for example, "mask_EGFR".
        
    Optional columns are:

    - "rotation_mode" : Random rotations mode used in analysis. Values
        must be one of {"3D", "2D", "None"}. Default: "2D".
    - "nn_plotted" : Number of nearest neighbors plotted in the NND. 
        Only integer values are accepted. Default: 4.
    
    When saving, each analysis run index is used as the prefix for 
    filename, for example, "analysis_run1_fit_summary.txt".

    Parameters
    ----------
    parameters_filename : str
        Path to the parameters file.
    asynch : bool (default=True)
        If True, multiprocessing is used.
    bootstrap : bool (default=False)
        If True, bootstrapping is used.
    verbose : bool (default=True)
        If True, progress bar for each row is printed to the console.
    """

    # print(f"asynch: {asynch}")
    # print(f"bootstrap: {bootstrap}")
    # print(f"verbose: {verbose}")

    import os
    import yaml
    from datetime import datetime
    import numpy as np
    import pandas as pd
    from . import io, spinna

    # open the parameters file
    if not isinstance(parameters_filename, str):
        raise TypeError("parameters_filename must be a string ending with .csv")
    elif not parameters_filename.endswith(".csv"):
        raise TypeError("parameters_filename must end with .csv")
    
    parameters = pd.read_csv(parameters_filename)

    # find the folder name for saving results
    result_dir = parameters_filename.replace(".csv", "_fitting_results")
    if os.path.isdir(result_dir): 
        i = 1
        while True:
            result_dir_ = result_dir + f"_{i}"
            if not os.path.isdir(result_dir_):
                result_dir = result_dir_
                break
            else:
                i += 1

    # check that all columns (non-target specific) are present
    for column in [
        "structures_filename",
        "granularity",
        "save_filename",
        "NND_bin",
        "NND_maxdist",
        "sim_repeats",
    ]:
        if column not in parameters.columns:
            raise ValueError(
                f"Column {column} not found in the parameters file."
            )

    # summary list of results for each simulation
    summary = []

    # run each row (analysis) one by one:
    for index, row in parameters.iterrows():
        print(f"Running SPINNA on row {index+1} out of {len(parameters)}.")
        # start by reading structures filename and creating the structures
        structures_filename = row["structures_filename"]
        structures, targets = spinna.load_structures(structures_filename)

        # get the target-independent parameters
        granularity = row["granularity"]
        NND_bin = row["NND_bin"]
        NND_maxdist = row["NND_maxdist"]
        sim_repeats = row["sim_repeats"]
        save_filename, _ = os.path.splitext(row["save_filename"])
        save_filename = os.path.join(result_dir, save_filename)

        # get the optional arguments
        random_rot_mode = "2D"
        if "rotation_mode" in row.index:
            if not isinstance(row["rotation_mode"], str):
                print("Invalid rotation_mode. Using default: 2D")
            else:
                random_rot_mode = str(row["rotation_mode"])

        nn_plotted = 4
        if "nn_plotted" in row.index:
            if not isinstance(row["nn_plotted"], int):
                print("Invalid nn_plotted. Using default: 4")
            else:
                nn_plotted = int(row["nn_plotted"])

        # initialize the input dictionaries that are target-specific
        label_unc = {}
        le = {}
        exp_data = {}
        n_simulated = {}

        # load data and parameters for each molecular target
        for target in targets:
            for col_name in [
                f"{_}_{target}" 
                for _ in ["label_unc", "le", "exp_data"]
            ]:
                if col_name not in row.index:
                    raise ValueError(
                        f"Column {col_name} not found in the parameters file."
                    )
            
            # load label uncertainy and labelling efficiency
            label_unc[target] = float(row[f"label_unc_{target}"])
            le[target] = float(row[f"le_{target}"]) / 100

            # load experimental data
            locs, info = io.load_locs(str(row[f"exp_data_{target}"]))
            pixelsize = 130
            for element in info:
                if "Picasso Localize" in element.values():
                    if "Pixelsize" in element:
                        pixelsize = element["Pixelsize"]
                        break
            if hasattr(locs, "z"):
                exp_data[target] = np.stack(
                    (locs.x * pixelsize, locs.y * pixelsize, locs.z)   
                ).T
                dim = 3
            else:
                exp_data[target] = np.stack(
                    (locs.x * pixelsize, locs.y * pixelsize)
                ).T
                dim = 2

            # number of simulated molecules (after labelling efficiency
            # correction)
            n_simulated[target] = int(len(locs) / le[target])


        # check if the distribution is homogeneous or heterogeneous
        apply_mask = True
        if dim == 3: # 3D simulation
            area = None
            if "volume" in row.index:
                volume = float(row["volume"])
                apply_mask = False
                # find z range
                if "z_range" not in row.index:
                    raise ValueError(
                        f"Column z_range not found in the parameters file."
                        " 3D simulation was specified with homogeneous"
                        " distribution. Please specify z_range."
                    )
                else:
                    z_range = float(row["z_range"])
        elif dim == 2: # 2D simulation
            volume = None
            if "area" in row.index:
                area = float(row["area"])
                apply_mask = False
        
        # extract masks if area/volume is not provided
        if apply_mask: 
            mask_paths = {}
            for target in targets:
                if f"mask_filename_{target}" not in row.index:
                    raise ValueError(
                        f"Column mask_filename_{target} not found in the"
                        " parameters file."
                    )
                else:
                    mask_paths[target] = row[f"mask_filename_{target}"]
            
        # generate search space for fitting
        N_structures = spinna.generate_N_structures(
            structures, n_simulated, granularity
        )

        # set up StructureMixer
        if apply_mask:
            masks = {}
            mask_info = {}
            width = height = depth = None
            for target in targets:
                masks[target] = np.load(mask_paths[target])
                mask_info[target] = yaml.load(
                    open(mask_paths[target].replace(".npy", ".yaml"), "r"),
                    Loader=yaml.FullLoader,
                )
            mask_dict = {"mask": masks, "info": mask_info}
        else:
            mask_dict = None
            if dim == 2:
                width = height = np.sqrt(area * 1e6)
                depth = None
            elif dim == 3:
                depth = z_range
                width = height = np.sqrt(volume * 1e9 / depth)

        mixer = spinna.StructureMixer(
            structures=structures,
            label_unc=label_unc,
            le=le,
            mask_dict=mask_dict,
            width=width, height=height, depth=depth,
            random_rot_mode=random_rot_mode,
        )
        
        if not os.path.isdir(result_dir): # create the save folder
            os.mkdir(result_dir)

        # set up and run fitting
        opt_props, score = spinna.SPINNA(
            mixer=mixer,
            gt_coords=exp_data,
            N_sim=sim_repeats,
        ).fit_stoichiometry(
            N_structures, 
            save=f"{save_filename}_fit_scores.csv", 
            asynch=asynch,
            bootstrap=bootstrap,
            callback="console" if verbose else None,
        )

        # save the results
        results = {}
        results["Date"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        results["File location of structures"] = structures_filename
        results["Molecular targets"] = targets
        results["File location of experimenal data"] = [
            str(row[f"exp_data_{target}"]) for target in targets
        ]
        results["Labelling efficiency (%)"] = [
            le[target] * 100 for target in targets
        ]
        results["Label uncertainty (nm)"] = list(label_unc.values())
        results["Rotation mode"] = random_rot_mode
        results["Dimensionality"] = f"{dim}D"
        results["Parameters search space granularity"] = granularity
        results["Fitted structures names"] = list(N_structures.keys())
        results["Number of simulation repeats"] = sim_repeats
        if isinstance(opt_props, tuple):
            props_mean, props_std = opt_props
            results["Modified Kolmogorov-Smirnov score +/- s.d."] = score
            results["Fitted proportions of structures"] = ", ".join([
                f"{props_mean[i]:.2f} +/- {props_std[i]:.2f}%" 
                for i in range(len(props_mean))
            ])
        else:
            results["Modified Kolmogorov-Smirnov score"] = score
            results["Fitted proportions of structures"] = opt_props
        
        # relative proportions of structures for each target
        if len(targets) > 1:
            for target in targets:
                rel_props = mixer.convert_props_for_target(
                    opt_props, target, n_simulated,
                )
                idx_valid = np.where(rel_props != np.inf)[0]
                value = ", ".join([
                    f"{structures[i].title}: {rel_props[i]:.2f}%" 
                    for i in idx_valid
                ])
                results[f"Relative proportions of {target} in"] = value
        
        if apply_mask:
            results["File location of masks"] = [
                row[f"mask_filename_{target}"] for target in targets
            ]
        else:
            if dim == 2:
                results["Area (um^2)"] = area
            elif dim == 3:
                results["Volume (um^3)"] = volume
                results["Z range (nm)"] = z_range
        # save .txt with summary of the results
        with open(f"{save_filename}_fit_summary.txt", "w") as f:
            for key, value in results.items():
                f.write(f"{key}: {value}\n")
        print(f"Results saved to {save_filename}_fit_summary.txt")
        summary.append(results)

        # plot and save the NND plots
        nn_counts = {}
        for i, t1 in enumerate(targets):
            for t2 in targets[i:]:
                nn_counts[f"{t1}-{t2}"] = nn_plotted
        mixer.nn_counts = nn_counts
        n_total = sum(n_simulated.values())
        if isinstance(opt_props, tuple):
            dist_sim = spinna.get_NN_dist_simulated(
                mixer.convert_props_to_counts(opt_props[0], n_total), 
                sim_repeats, 
                mixer, 
                duplicate=True,
            )
        else:
            dist_sim = spinna.get_NN_dist_simulated(
                mixer.convert_props_to_counts(opt_props, n_total), 
                sim_repeats, 
                mixer, 
                duplicate=True,
            )
        for i, (t1, t2, _) in enumerate(mixer.get_neighbor_idx(duplicate=True)):
            fig, ax = spinna.plot_NN(
                dist=dist_sim[i], mode='plot', show_legend=False,
                return_fig=True, figsize=(4.947, 3.71), alpha=1.0,
                binsize=NND_bin, xlim=[0, NND_maxdist],
                title=f"Nearest Neighbors Distances: {t1} -> {t2}",
            )
            exp1 = exp_data[t1]
            exp2 = exp_data[t2]
            fig, ax = spinna.plot_NN(
                data1=exp1, data2=exp2, 
                n_neighbors=nn_plotted, 
                show_legend=False, fig=fig, ax=ax, mode='hist', 
                return_fig=True, binsize=NND_bin, xlim=[0, NND_maxdist],
                title=f"Nearest Neighbors Distances: {t1} -> {t2}",
                savefig=[
                    f"{save_filename}_NND_{t1}_{t2}.{_}" 
                    for _ in ["png", "svg"]
                ],
            ) 

    # save the summary as .csv file
    summary = pd.DataFrame(summary)
    summary.to_csv(os.path.join(result_dir, "summary_results.csv"), index=False)


def main():

    # Main parser
    parser = argparse.ArgumentParser("picasso")
    subparsers = parser.add_subparsers(dest="command")

    for command in ["toraw", "filter"]:
        subparsers.add_parser(command)

    # link parser
    link_parser = subparsers.add_parser(
        "link", help="link localizations in consecutive frames"
    )
    link_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    link_parser.add_argument(
        "-d",
        "--distance",
        type=float,
        default=1.0,
        help=(
            "maximum distance between localizations"
            " to consider them the same binding event (default=1.0)"
        ),
    )
    link_parser.add_argument(
        "-t",
        "--tolerance",
        type=int,
        default=1,
        help=(
            "maximum dark time between localizations"
            " to still consider them the same binding event (default=1)"
        ),
    )

    cluster_combine_parser = subparsers.add_parser(
        "cluster_combine",
        help="combine localization in each cluster of a group",
    )
    cluster_combine_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )

    cluster_combine_dist_parser = subparsers.add_parser(
        "cluster_combine_dist",
        help="calculate the nearest neighbor for each combined cluster",
    )
    cluster_combine_dist_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )

    clusterfilter_parser = subparsers.add_parser(
        "clusterfilter",
        help="filter localizations by properties of their clusters",
    )
    clusterfilter_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    clusterfilter_parser.add_argument("clusterfile", help="a hdf5 clusterfile")
    clusterfilter_parser.add_argument(
        "parameter", type=str, help="parameter to be filtered"
    )
    clusterfilter_parser.add_argument("minval", type=float, help="lower boundary")
    clusterfilter_parser.add_argument("maxval", type=float, help="upper boundary")

    # undrift parser
    undrift_parser = subparsers.add_parser(
        "undrift", help="correct localization coordinates for drift"
    )
    undrift_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    # undrift_parser.add_argument(
    #     "-m",
    #     "--mode",
    #     default="render",
    #     help='"std", "render" or "framepair")',
    # )
    undrift_parser.add_argument(
        "-s",
        "--segmentation",
        type=float,
        default=1000,
        help=(
            "the number of frames to be combined"
            " for one temporal segment (default=1000)"
        ),
    )
    undrift_parser.add_argument(
        "-f",
        "--fromfile",
        type=str,
        help="apply drift from specified file instead of computing it",
    )
    undrift_parser.add_argument(
        "-d",
        "--nodisplay",
        action="store_false",
        help="do not display estimated drift",
    )

    # undrift by AIM parser
    undrift_aim_parser = subparsers.add_parser(
        "aim", help="correct localization coordinates for drift with AIM"
    )
    undrift_aim_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    undrift_aim_parser.add_argument(
        "-s",
        "--segmentation",
        type=float,
        default=100,
        help=(
            "the number of frames to be combined"
            " for one temporal segment (default=100)"
        ),
    )
    undrift_aim_parser.add_argument(
        "-i",
        "--intersectdist",
        type=float,
        default=20/130,
        help=(
            "max. distance (cam. pixels) between localizations in"
            " consecutive segments to be considered as intersecting"
        ),
    )
    undrift_aim_parser.add_argument(
        "-r",
        "--roiradius",
        type=float,
        default=60/130,
        help=(
            "max. drift (cam. pixels) between two consecutive"
            " segments"
        ),
    )

    # local densitydd
    density_parser = subparsers.add_parser(
        "density", help="compute the local density of localizations"
    )
    density_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    density_parser.add_argument(
        "radius",
        type=float,
        help=("maximal distance between to localizations" " to be considered local"),
    )

    # DBSCAN
    dbscan_parser = subparsers.add_parser(
        "dbscan",
        help="cluster localizations with the dbscan clustering algorithm",
    )
    dbscan_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    dbscan_parser.add_argument(
        "radius",
        type=float,
        help=("maximal distance (camera pixels) between to localizations" " to be considered local"),
    )
    dbscan_parser.add_argument(
        "density",
        type=int,
        help=("minimum local density for localizations" " to be assigned to a cluster"),
    )
    dbscan_parser.add_argument(
        "pixelsize",
        type=int,
        help=("camera pixel size in nm (required for 3D localizations only)"),
        default=None,
    )

    # HDBSCAN
    hdbscan_parser = subparsers.add_parser(
        "hdbscan",
        help="cluster localizations with the hdbscan clustering algorithm",
    )
    hdbscan_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    hdbscan_parser.add_argument(
        "min_cluster",
        type=int,
        help=("smallest size grouping that is considered a cluster"),
    )
    hdbscan_parser.add_argument(
        "min_samples",
        type=int,
        help=("the higher the more points are considered noise"),
    )
    hdbscan_parser.add_argument(
        "pixelsize",
        type=int,
        help=("camera pixel size in nm (required for 3D localizations only)"),
        default=None,
    )

    # SMLM clusterer
    smlm_cluster_parser = subparsers.add_parser(
        "smlm_cluster",
        help="cluster localizations with the custom SMLM clustering algorithm",
    )
    smlm_cluster_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )
    smlm_cluster_parser.add_argument(
        "radius",
        type=float,
        help=("clustering radius (in camera pixels)"),
    )
    smlm_cluster_parser.add_argument(
        "min_locs",
        type=int,
        help=("minimum number of localizations in a cluster"),
    )
    smlm_cluster_parser.add_argument(
        "pixelsize",
        type=int,
        help=("camera pixel size in nm (required for 3D localizations only)"),
        default=None,
    )
    smlm_cluster_parser.add_argument(
        "basic_fa",
        type=bool,
        help=("whether or not perform basic frame analysis (sticking event removal)"),
        default=False,
    )
    smlm_cluster_parser.add_argument(
        "radius_z",
        type=float,
        help=(
            "clustering radius in axial direction (MUST BE SET FOR 3D!!!)"
        ),
        default=None,
    )

    # Dark time
    dark_parser = subparsers.add_parser(
        "dark", help="compute the dark time for grouped localizations"
    )
    dark_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )

    # align
    align_parser = subparsers.add_parser(
        "align", help="align one localization file to another"
    )
    align_parser.add_argument(
        "-d", "--display", help="display correlation", action="store_true"
    )
    # align_parser.add_argument('-a', '--affine',
    # help='include affine transformations (may take long time)',
    # action='store_true')
    align_parser.add_argument(
        "file", help="one or multiple hdf5 localization files", nargs="+"
    )

    # join
    join_parser = subparsers.add_parser(
        "join",
        help="join hdf5 localization lists. frame numbers of consecutive files will be reindexed.",
    )
    join_parser.add_argument(
        "file", nargs="+", help="the hdf5 localization files to be joined"
    )
    join_parser.add_argument(
        "-k", "--keepindex", help="do not change frame numbers", action="store_true"
    )

    # group properties
    groupprops_parser = subparsers.add_parser(
        "groupprops",
        help=("calculate kinetics " "and various properties of localization groups"),
    )
    groupprops_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            " specified by a unix style path pattern"
        ),
    )

    # Pair correlation
    pc_parser = subparsers.add_parser(
        "pc", help="calculate the pair-correlation of localizations"
    )
    pc_parser.add_argument(
        "-b", "--binsize", type=float, default=0.1, help="the bin size"
    )
    pc_parser.add_argument(
        "-r",
        "--rmax",
        type=float,
        default=10,
        help="The maximum distance to calculate the pair-correlation",
    )
    pc_parser.add_argument(
        "files",
        help=(
            "one or multiple hdf5 localization files"
            "specified by a unix style path pattern"
        ),
    )

    # localize
    localize_parser = subparsers.add_parser(
        "localize", help="identify and fit single molecule spots"
    )
    localize_parser.add_argument(
        "files",
        nargs="?",
        help=(
            "one movie file or a folder containing movie files"
            " specified by a unix style path pattern"
        ),
    )
    localize_parser.add_argument(
        "-b", "--box-side-length", type=int, default=7, help="box side length"
    )
    localize_parser.add_argument(
        "-a",
        "--fit-method",
        choices=["mle", "lq", "lq-gpu", "lq-3d", "lq-gpu-3d", "avg"],
        default="mle",
    )
    localize_parser.add_argument(
        "-g", "--gradient", type=int, default=5000, help="minimum net gradient"
    )
    localize_parser.add_argument(
        "-d",
        "--drift",
        type=int,
        default=1000,
        help="segmentation size for subsequent RCC, 0 to deactivate",
    )
    localize_parser.add_argument(
        "-r",
        "--roi",
        type=int,
        nargs=4,
        default=None,
        help=(
            "ROI (y_min, x_min, y_max, x_max) in camera pixels;\n"
            "note the origin of the image is in the top left corner"
        ),
    )
    localize_parser.add_argument(
        "-bl", "--baseline", type=int, default=0, help="camera baseline"
    )
    localize_parser.add_argument(
        "-s", "--sensitivity", type=float, default=1, help="camera sensitivity"
    )
    localize_parser.add_argument(
        "-ga", "--gain", type=int, default=1, help="camera gain"
    )
    localize_parser.add_argument(
        "-qe", "--qe", type=float, default=1, help="camera quantum efficiency"
    )
    localize_parser.add_argument(
        "-mf", "--mf", type=float, default=0, help="Magnification factor (only 3d)"
    )
    localize_parser.add_argument(
        "-px", "--pixelsize", type=int, default=130, help="pixelsize in nm"
    )
    localize_parser.add_argument(
        "-zc",
        "--zc",
        type=str,
        default="",
        help="Path to 3d calibration file (only 3d)",
    )

    localize_parser.add_argument(
        "-sf",
        "--suffix",
        type=str,
        default="",
        help="Suffix to add to files",
    )
    
    localize_parser.add_argument(
        "-db",
        "--database",
        action="store_true",
        help="do not add to database",
    )

    # nneighbors
    nneighbor_parser = subparsers.add_parser(
        "nneighbor", help="calculate nearest neighbor of a clustered dataset"
    )
    nneighbor_parser.add_argument(
        "files",
        nargs="?",
        help=(
            "one or multiple hdf5 clustered files"
            " specified by a unix style path pattern"
        ),
    )

    # render
    render_parser = subparsers.add_parser(
        "render", help="render localization based images"
    )
    render_parser.add_argument(
        "files",
        nargs="?",
        help=(
            "one or multiple localization files"
            " specified by a unix style path pattern"
        ),
    )
    render_parser.add_argument(
        "-o",
        "--oversampling",
        type=float,
        default=1.0,
        help="the number of super-resolution pixels per camera pixels",
    )
    render_parser.add_argument(
        "-b",
        "--blur-method",
        choices=["none", "convolve", "gaussian"],
        default="convolve",
    )
    render_parser.add_argument(
        "-w",
        "--min-blur-width",
        type=float,
        default=0.0,
        help="minimum blur width if blur is applied",
    )
    render_parser.add_argument(
        "--vmin",
        type=float,
        default=0.0,
        help="minimum colormap level in range 0-100 or absolute value",
    )
    render_parser.add_argument(
        "--vmax",
        type=float,
        default=20.0,
        help="maximum colormap level in range 0-100 or absolute value",
    )
    render_parser.add_argument(
        "--scaling",
        choices=["yes", "no"],
        default="yes",
        help="if scaling the colormap value is relative in the range 0-100",
    )
    render_parser.add_argument(
        "-c",
        "--cmap",
        choices=["viridis", "inferno", "plasma", "magma", "hot", "gray"],
        help="the colormap to be applied",
    )
    render_parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        help="do not open the image file",
    )

    # design
    subparsers.add_parser("design", help="design RRO DNA origami structures")
    # simulate
    subparsers.add_parser("simulate", help="simulate single molecule fluorescence data")

    # nanotron
    nanotron_parser = subparsers.add_parser(
        "nanotron", help="segmentation with deep learning"
    )
    nanotron_parser.add_argument(
        "-m",
        "--model",
        nargs="?",
        help="a model file for prediction",
    )

    nanotron_parser.add_argument(
        "files",
        nargs="?",
        help="one localization file or a folder containing localization files"
        " specified by a unix style path pattern",
    )

    # average
    average_parser = subparsers.add_parser("average", help="particle averaging")
    average_parser.add_argument(
        "-o",
        "--oversampling",
        type=float,
        default=10,
        help=(
            "oversampling of the super-resolution images" " for alignment evaluation"
        ),
    )
    average_parser.add_argument("-i", "--iterations", type=int, default=20)
    average_parser.add_argument(
        "files",
        nargs="?",
        help="a localization file with grouped localizations",
    )

    average3_parser = subparsers.add_parser(
        "average3", help="three-dimensional particle averaging"
    )

    hdf2visp_parser = subparsers.add_parser("hdf2visp")
    hdf2visp_parser.add_argument("files")
    hdf2visp_parser.add_argument("pixelsize", type=float)

    csv2hdf_parser = subparsers.add_parser("csv2hdf")
    csv2hdf_parser.add_argument("files")
    csv2hdf_parser.add_argument("pixelsize", type=float)

    hdf2csv_parser = subparsers.add_parser("hdf2csv")
    hdf2csv_parser.add_argument("files")

    server_parser = subparsers.add_parser(
        "server", help="picasso server workflow management system"
    )

    spinna_parser = subparsers.add_parser(
        "spinna",
        help=(
            "picasso single protein investigation via nearest neighbor analysis"
        ),
    )
    spinna_parser.add_argument(
        "-p",
        "--parameters",
        type=str,
        help=(
            ".csv file containing the parameters for spinna batch analysis;"
            " see the documentation for details explaining the .csv file"
            " structure."
        ),
    )
    spinna_parser.add_argument(
        "-a",
        "--asynch",
        action="store_false",
        help="do not perform fitting asynchronously (multiprocessing)",
    )
    spinna_parser.add_argument(
        "-b",
        "--bootstrap",
        action="store_true",
        help="perform bootstrapping",
    )
    spinna_parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="display progress bar for each row",
    )

    # Parse
    args = parser.parse_args()
    if args.command:
        if args.command == "toraw":
            from .gui import toraw

            toraw.main()
        elif args.command == "localize":
            if args.files:
                _localize(args)
            else:
                from picasso.gui import localize

                localize.main()
        elif args.command == "filter":
            from .gui import filter

            filter.main()
        elif args.command == "render":
            if args.files:
                _render(args)
            else:
                from .gui import render

                render.main()
        elif args.command == "average":
            if args.files:
                _average(args)
            else:
                from .gui import average

                average.main()
        elif args.command == "nanotron":
            if args.files:
                _nanotron(args)
            else:
                from .gui import nanotron

                nanotron.main()
        elif args.command == "average3":
            from .gui import average3

            average3.main()
        elif args.command == "link":
            _link(args.files, args.distance, args.tolerance)
        elif args.command == "cluster_combine":
            _cluster_combine(args.files)
        elif args.command == "cluster_combine_dist":
            _cluster_combine_dist(args.files)
        elif args.command == "clusterfilter":
            _clusterfilter(
                args.files,
                args.clusterfile,
                args.parameter,
                args.minval,
                args.maxval,
            )
        elif args.command == "undrift":
            _undrift(args.files, args.segmentation, args.nodisplay, args.fromfile)
        elif args.command == "aim":
            _undrift_aim(args.files, args.segmentation, args.intersectdist, args.roiradius)
        elif args.command == "density":
            _density(args.files, args.radius)
        elif args.command == "dbscan":
            _dbscan(args.files, args.radius, args.density, args.pixelsize)
        elif args.command == "hdbscan":
            _hdbscan(args.files, args.min_cluster, args.min_samples, args.pixelsize)
        elif args.command == "smlm_cluster":
            _smlm_clusterer(
                args.files,
                args.radius,
                args.min_locs,
                args.pixelsize,
                args.basic_fa,
                args.radius_z,
            )
        elif args.command == "nneighbor":
            _nneighbor(args.files)
        elif args.command == "dark":
            _dark(args.files)
        elif args.command == "align":
            _align(args.file, args.display)
        elif args.command == "join":
            _join(args.file, args.keepindex)
        elif args.command == "groupprops":
            _groupprops(args.files)
        elif args.command == "pc":
            _pair_correlation(args.files, args.binsize, args.rmax)
        elif args.command == "simulate":
            from .gui import simulate

            simulate.main()
        elif args.command == "design":
            from .gui import design

            design.main()
        elif args.command == "hdf2visp":
            _hdf2visp(args.files, args.pixelsize)
        elif args.command == "csv2hdf":
            _csv2hdf(args.files, args.pixelsize)
        elif args.command == "hdf2csv":
            _hdf2csv(args.files)
        elif args.command == "server":
            _start_server()
        elif args.command == "spinna":
            if args.parameters:
                _spinna_batch_analysis(args.parameters, args.asynch, args.bootstrap)
            else:
                from .gui import spinna

                spinna.main()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()
